import functools
import random
from itertools import chain, count, cycle

import autoray as ar

from quimb.gen.operators import swap
from quimb.tensor.tensor_1d import maybe_factor_gate_into_tensor
from quimb.tensor.tensor_core import (
    Tensor,
    bonds,
    rand_uuid,
    tags_to_oset,
    tensor_contract,
)
from quimb.utils import pairwise


def manhattan_distance(coo_a, coo_b):
    return sum(abs(coo_a[i] - coo_b[i]) for i in range(2))


def gate_string_split_(
    TG,
    where,
    string,
    original_ts,
    bonds_along,
    reindex_map,
    site_ix,
    info,
    **compress_opts,
):
    # by default this means singuvalues are kept in the string 'blob' tensor
    compress_opts.setdefault("absorb", "right")

    # the outer, neighboring indices of each tensor in the string
    neighb_inds = []

    # tensors we are going to contract in the blob, reindex some to attach gate
    contract_ts = []

    for t, coo in zip(original_ts, string):
        neighb_inds.append(tuple(ix for ix in t.inds if ix not in bonds_along))
        contract_ts.append(t.reindex(reindex_map) if coo in where else t)

    # form the central blob of all sites and gate contracted
    blob = tensor_contract(*contract_ts, TG)

    regauged = []

    # one by one extract the site tensors again from each end
    inner_ts = [None] * len(string)
    i = 0
    j = len(string) - 1

    while True:
        lix = neighb_inds[i]
        if i > 0:
            lix += (bonds_along[i - 1],)

        # the original bond we are restoring
        bix = bonds_along[i]

        # split the blob!
        inner_ts[i], *maybe_svals, blob = blob.split(
            left_inds=lix, get="tensors", bond_ind=bix, **compress_opts
        )

        # if singular values are returned (``absorb=None``) check if we should
        #     return them via ``info``, e.g. for ``SimpleUpdate`
        if maybe_svals and info is not None:
            s = next(iter(maybe_svals)).data
            coo_pair = tuple(sorted((string[i], string[i + 1])))
            info["singular_values", coo_pair] = s

            # regauge the blob but record so as to unguage later
            if i != j - 1:
                blob.multiply_index_diagonal_(bix, s)
                regauged.append((i + 1, bix, s))

        # move inwards along string, terminate if two ends meet
        i += 1
        if i == j:
            inner_ts[i] = blob
            break

        # extract at end of string
        lix = neighb_inds[j]
        if j < len(string) - 1:
            lix += (bonds_along[j],)

        # the original bond we are restoring
        bix = bonds_along[j - 1]

        # split the blob!
        inner_ts[j], *maybe_svals, blob = blob.split(
            left_inds=lix, get="tensors", bond_ind=bix, **compress_opts
        )

        # if singular values are returned (``absorb=None``) check if we should
        #     return them via ``info``, e.g. for ``SimpleUpdate`
        if maybe_svals and info is not None:
            s = next(iter(maybe_svals)).data
            coo_pair = tuple(sorted((string[j - 1], string[j])))
            info["singular_values", coo_pair] = s

            # regauge the blob but record so as to unguage later
            if j != i + 1:
                blob.multiply_index_diagonal_(bix, s)
                regauged.append((j - 1, bix, s))

        # move inwards along string, terminate if two ends meet
        j -= 1
        if j == i:
            inner_ts[j] = blob
            break

    # ungauge the site tensors along bond if necessary
    for i, bix, s in regauged:
        t = inner_ts[i]
        t.multiply_index_diagonal_(bix, s**-1)

    # transpose to match original tensors and update original data
    for to, tn in zip(original_ts, inner_ts):
        tn.transpose_like_(to)
        to.modify(data=tn.data)


def gate_string_reduce_split_(
    TG,
    where,
    string,
    original_ts,
    bonds_along,
    reindex_map,
    site_ix,
    info,
    **compress_opts,
):
    # by default this means singuvalues are kept in the string 'blob' tensor
    compress_opts.setdefault("absorb", "right")

    # indices to reduce, first and final include physical indices for gate
    inds_to_reduce = [(bonds_along[0], site_ix[0])]
    for b1, b2 in pairwise(bonds_along):
        inds_to_reduce.append((b1, b2))
    inds_to_reduce.append((bonds_along[-1], site_ix[-1]))

    # tensors that remain on the string sites and those pulled into string
    outer_ts, inner_ts = [], []
    for coo, rix, t in zip(string, inds_to_reduce, original_ts):
        tq, tr = t.split(
            left_inds=None, right_inds=rix, method="qr", get="tensors"
        )
        outer_ts.append(tq)
        inner_ts.append(tr.reindex_(reindex_map) if coo in where else tr)

    # contract the blob of gate with reduced tensors only
    blob = tensor_contract(*inner_ts, TG)

    regauged = []

    # extract the new reduced tensors sequentially from each end
    i = 0
    j = len(string) - 1

    while True:
        # extract at beginning of string
        lix = bonds(blob, outer_ts[i])
        if i == 0:
            lix.add(site_ix[0])
        else:
            lix.add(bonds_along[i - 1])

        # the original bond we are restoring
        bix = bonds_along[i]

        # split the blob!
        inner_ts[i], *maybe_svals, blob = blob.split(
            left_inds=lix, get="tensors", bond_ind=bix, **compress_opts
        )

        # if singular values are returned (``absorb=None``) check if we should
        #     return them via ``info``, e.g. for ``SimpleUpdate`
        if maybe_svals and info is not None:
            s = next(iter(maybe_svals)).data
            coo_pair = tuple(sorted((string[i], string[i + 1])))
            info["singular_values", coo_pair] = s

            # regauge the blob but record so as to unguage later
            if i != j - 1:
                blob.multiply_index_diagonal_(bix, s)
                regauged.append((i + 1, bix, s))

        # move inwards along string, terminate if two ends meet
        i += 1
        if i == j:
            inner_ts[i] = blob
            break

        # extract at end of string
        lix = bonds(blob, outer_ts[j])
        if j == len(string) - 1:
            lix.add(site_ix[-1])
        else:
            lix.add(bonds_along[j])

        # the original bond we are restoring
        bix = bonds_along[j - 1]

        # split the blob!
        inner_ts[j], *maybe_svals, blob = blob.split(
            left_inds=lix, get="tensors", bond_ind=bix, **compress_opts
        )

        # if singular values are returned (``absorb=None``) check if we should
        #     return them via ``info``, e.g. for ``SimpleUpdate`
        if maybe_svals and info is not None:
            s = next(iter(maybe_svals)).data
            coo_pair = tuple(sorted((string[j - 1], string[j])))
            info["singular_values", coo_pair] = s

            # regauge the blob but record so as to unguage later
            if j != i + 1:
                blob.multiply_index_diagonal_(bix, s)
                regauged.append((j - 1, bix, s))

        # move inwards along string, terminate if two ends meet
        j -= 1
        if j == i:
            inner_ts[j] = blob
            break

    # reabsorb the inner reduced tensors into the sites
    new_ts = [
        tensor_contract(ts, tr, output_inds=to.inds)
        for to, ts, tr in zip(original_ts, outer_ts, inner_ts)
    ]

    # ungauge the site tensors along bond if necessary
    for i, bix, s in regauged:
        t = new_ts[i]
        t.multiply_index_diagonal_(bix, s**-1)

    # update originals
    for to, t in zip(original_ts, new_ts):
        to.modify(data=t.data)


def gate_2d_long_range(
    self,
    G,
    where,
    contract=False,
    tags=None,
    inplace=False,
    info=None,
    long_range_use_swaps=False,
    long_range_path_sequence=None,
    **compress_opts,
):
    psi = self if inplace else self.copy()

    ng = len(where)

    dp = psi.phys_dim(*where[0])
    tags = tags_to_oset(tags)

    # allow a matrix to be reshaped into a tensor if it factorizes
    #     i.e. (4, 4) assumed to be two qubit gate -> (2, 2, 2, 2)
    G = maybe_factor_gate_into_tensor(G, dp, ng, where)

    site_ix = [psi.site_ind(i, j) for i, j in where]
    # new indices to join old physical sites to new gate
    bnds = [rand_uuid() for _ in range(ng)]
    reindex_map = dict(zip(site_ix, bnds))

    TG = Tensor(G, inds=site_ix + bnds, tags=tags, left_inds=bnds)

    # following are all based on splitting tensors to maintain structure
    ij_a, ij_b = where

    # parse the argument specifying how to find the path between
    # non-nearest neighbours
    if long_range_path_sequence is not None:
        # make sure we can index
        long_range_path_sequence = tuple(long_range_path_sequence)
        # if the first element is a str specifying move sequence, e.g.
        #     ('v', 'h')
        #     ('av', 'bv', 'ah', 'bh')  # using swaps
        manual_lr_path = not isinstance(long_range_path_sequence[0], str)
        # otherwise assume a path has been manually specified, e.g.
        #     ((1, 2), (2, 2), (2, 3), ... )
        #     (((1, 1), (1, 2)), ((4, 3), (3, 3)), ...)  # using swaps
    else:
        manual_lr_path = False

    # check if we are not nearest neighbour and need to swap first
    if long_range_use_swaps:
        if manual_lr_path:
            *swaps, final = long_range_path_sequence
        else:
            # find a swap path
            *swaps, final = gen_long_range_swap_path(
                ij_a, ij_b, sequence=long_range_path_sequence
            )

        # move the sites together
        SWAP = get_swap(
            dp, dtype=ar.get_dtype_name(G), backend=ar.infer_backend(G)
        )
        for pair in swaps:
            psi.gate_(SWAP, pair, contract=contract, absorb="right")

        compress_opts["info"] = info
        compress_opts["contract"] = contract

        # perform actual gate also compressing etc on 'way back'
        psi.gate_(G, final, **compress_opts)

        compress_opts.setdefault("absorb", "both")
        for pair in reversed(swaps):
            psi.gate_(SWAP, pair, **compress_opts)

        return psi

    if manual_lr_path:
        string = long_range_path_sequence
    else:
        string = tuple(
            gen_long_range_path(*where, sequence=long_range_path_sequence)
        )

    # the tensors along this string, which will be updated
    original_ts = [psi[coo] for coo in string]

    # the len(string) - 1 indices connecting the string
    bonds_along = [
        next(iter(bonds(t1, t2))) for t1, t2 in pairwise(original_ts)
    ]

    if contract == "split":
        #
        #       │╱  │╱          │╱  │╱
        #     ──GGGGG──  ==>  ──G┄┄┄G──
        #      ╱   ╱           ╱   ╱
        #
        gate_string_split_(
            TG,
            where,
            string,
            original_ts,
            bonds_along,
            reindex_map,
            site_ix,
            info,
            **compress_opts,
        )

    elif contract == "reduce-split":
        #
        #       │   │             │ │
        #       GGGGG             GGG               │ │
        #       │╱  │╱   ==>     ╱│ │  ╱   ==>     ╱│ │  ╱          │╱  │╱
        #     ──●───●──       ──>─●─●─<──       ──>─GGG─<──  ==>  ──G┄┄┄G──
        #      ╱   ╱           ╱     ╱           ╱     ╱           ╱   ╱
        #    <QR> <LQ>                            <SVD>
        #
        gate_string_reduce_split_(
            TG,
            where,
            string,
            original_ts,
            bonds_along,
            reindex_map,
            site_ix,
            info,
            **compress_opts,
        )

    return psi


def gen_long_range_path(ij_a, ij_b, sequence=None):
    """Generate a string of coordinates, in order, from ``ij_a`` to ``ij_b``.

    Parameters
    ----------
    ij_a : (int, int)
        Coordinate of site 'a'.
    ij_b : (int, int)
        Coordinate of site 'b'.
    sequence : None, iterable of {'v', 'h'}, or 'random', optional
        What order to cycle through and try and perform moves in, 'v', 'h'
        standing for move vertically and horizontally respectively. The default
        is ``('v', 'h')``.

    Returns
    -------
    generator[tuple[int]]
        The path, each element is a single coordinate.
    """
    ia, ja = ij_a
    ib, jb = ij_b
    di = ib - ia
    dj = jb - ja

    # nearest neighbour
    if abs(di) + abs(dj) == 1:
        yield ij_a
        yield ij_b
        return

    if sequence is None:
        poss_moves = cycle(("v", "h"))
    elif sequence == "random":
        poss_moves = (random.choice("vh") for _ in count())
    else:
        poss_moves = cycle(sequence)

    yield ij_a

    for move in poss_moves:
        if abs(di) + abs(dj) == 1:
            yield ij_b
            return

        if (move == "v") and (di != 0):
            # move a vertically
            istep = min(max(di, -1), +1)
            new_ij_a = (ia + istep, ja)
            yield new_ij_a
            ij_a = new_ij_a
            ia += istep
            di -= istep
        elif (move == "h") and (dj != 0):
            # move a horizontally
            jstep = min(max(dj, -1), +1)
            new_ij_a = (ia, ja + jstep)
            yield new_ij_a
            ij_a = new_ij_a
            ja += jstep
            dj -= jstep


def gen_long_range_swap_path(ij_a, ij_b, sequence=None):
    """Generate the coordinates or a series of swaps that would bring ``ij_a``
    and ``ij_b`` together.

    Parameters
    ----------
    ij_a : (int, int)
        Coordinate of site 'a'.
    ij_b : (int, int)
        Coordinate of site 'b'.
    sequence : None, it of {'av', 'bv', 'ah', 'bh'}, or 'random', optional
        What order to cycle through and try and perform moves in, 'av', 'bv',
        'ah', 'bh' standing for move 'a' vertically, 'b' vertically, 'a'
        horizontally', and 'b' horizontally respectively. The default is
        ``('av', 'bv', 'ah', 'bh')``.

    Returns
    -------
    generator[tuple[tuple[int]]]
        The path, each element is two coordinates to swap.
    """
    ia, ja = ij_a
    ib, jb = ij_b
    di = ib - ia
    dj = jb - ja

    # nearest neighbour
    if abs(di) + abs(dj) == 1:
        yield (ij_a, ij_b)
        return

    if sequence is None:
        poss_moves = cycle(("av", "bv", "ah", "bh"))
    elif sequence == "random":
        poss_moves = (random.choice(("av", "bv", "ah", "bh")) for _ in count())
    else:
        poss_moves = cycle(sequence)

    for move in poss_moves:
        if (move == "av") and (di != 0):
            # move a vertically
            istep = min(max(di, -1), +1)
            new_ij_a = (ia + istep, ja)
            yield (ij_a, new_ij_a)
            ij_a = new_ij_a
            ia += istep
            di -= istep

        elif (move == "bv") and (di != 0):
            # move b vertically
            istep = min(max(di, -1), +1)
            new_ij_b = (ib - istep, jb)
            # need to make sure final gate is applied correct way
            if new_ij_b == ij_a:
                yield (ij_a, ij_b)
            else:
                yield (ij_b, new_ij_b)
            ij_b = new_ij_b
            ib -= istep
            di -= istep

        elif (move == "ah") and (dj != 0):
            # move a horizontally
            jstep = min(max(dj, -1), +1)
            new_ij_a = (ia, ja + jstep)
            yield (ij_a, new_ij_a)
            ij_a = new_ij_a
            ja += jstep
            dj -= jstep

        elif (move == "bh") and (dj != 0):
            # move b horizontally
            jstep = min(max(dj, -1), +1)
            new_ij_b = (ib, jb - jstep)
            # need to make sure final gate is applied correct way
            if new_ij_b == ij_a:
                yield (ij_a, ij_b)
            else:
                yield (ij_b, new_ij_b)
            ij_b = new_ij_b
            jb -= jstep
            dj -= jstep

        if di == dj == 0:
            return


def swap_path_to_long_range_path(swap_path, ij_a):
    """Generates the ordered long-range path - a sequence of coordinates - from
    a (long-range) swap path - a sequence of coordinate pairs.
    """
    sites = set(chain(*swap_path))
    return sorted(sites, key=lambda ij_b: manhattan_distance(ij_a, ij_b))


@functools.lru_cache(8)
def get_swap(dp, dtype, backend):
    SWAP = swap(dp, dtype=dtype)
    return ar.do("array", SWAP, like=backend)



"""Given a single definition of hilbert space and set of terms, build out
various representations of the operator.

This includes:

- Sparse matrix in various formats (CSR, CSC, COO, etc.)
- Dense matrix
- Matrix product operator (MPO) for DMRG etc.
- Dict of k-local terms (for use with PEPS etc.)
- Coupling function, for use with VMC etc.
"""

import functools
import operator

import numpy as np

from . import configcore
from .hilbertspace import HilbertSpace

_OPMAP = {
    "I": {0: (0, 1.0), 1: (1, 1.0)},
    # pauli matrices
    "x": {0: (1, 1.0), 1: (0, 1.0)},
    "y": {0: (1, 1.0j), 1: (0, -1.0j)},
    "z": {0: (0, 1.0), 1: (1, -1.0)},
    # ZX=iY: 'real Y'
    "ⴵ": {0: (1, -1.0), 1: (0, 1.0)},
    # spin 1/2 matrices (scaled paulis)
    "sx": {0: (1, 0.5), 1: (0, 0.5)},
    "sy": {0: (1, 0.5j), 1: (0, -0.5j)},
    "sz": {0: (0, 0.5), 1: (1, -0.5)},
    # creation / annihilation operators
    "+": {0: (1, 1.0)},
    "-": {1: (0, 1.0)},
    # number, symmetric number, and hole operators
    "n": {1: (1, 1.0)},
    "sn": {0: (0, -0.5), 1: (1, 0.5)},
    "h": {0: (0, 1.0)},
}

_OPCOMPLEX = {
    op
    for op, mat in _OPMAP.items()
    if any(isinstance(coeff, complex) for _, (_, coeff) in mat.items())
}


@functools.lru_cache(maxsize=None)
def get_mat(op, dtype=None):
    if dtype is None:
        if any(
            isinstance(coeff, complex) for _, (_, coeff) in _OPMAP[op].items()
        ):
            dtype = np.complex128
        else:
            dtype = np.float64

    a = np.zeros((2, 2), dtype=dtype)
    for j, (i, xij) in _OPMAP[op].items():
        a[i, j] = xij
    # make immutable since caching
    a.flags.writeable = False
    return a


def _identity_fn(x):
    return x


def jordan_wigner_transform(terms, site_to_reg=None, reg_to_site=None):
    """Transform the terms in this operator by pre-prending pauli Z
    strings to all creation and annihilation operators. This is always
    applied directly to the raw terms, so that the any fermionic ordering
    is respected. Any further transformations (e.g. simplification or
    pauli decomposition) should thus be applied after this transformation.

    Note this doesn't decompose +, - into (X + iY) / 2 and (X - iY) / 2, it
    just prepends Z strings. Call `pauli_decompose` after this to get the
    full decomposition.

    The placement of the Z strings is defined by the ordering supplied by
    `site_to_reg` and `reg_to_site`, by default, it assumes the terms already
    are specified on a linear range of integers.

    Parameters
    ----------
    terms : dict[term, coeff]
        The terms of the operator. Each term is a tuple of tuples, where each
        inner tuple is a pair of ``(operator, site)``.
    site_to_reg : callable, optional
        A function that maps a site to a linear register index. If not
        provided, the sites are assumed to be linear integers already.
    reg_to_site : callable, optional
        A function that maps a linear register index to a site. If not
        provided, the sites are assumed to be linear integers already.

    Returns
    -------
    terms_jordan_wigner : dict[term, coeff]
        The transformed terms of the operator. Each term is a tuple of tuples,
        where each inner tuple is a pair of ``(operator, site)``.
    """
    if site_to_reg is None:
        site_to_reg = _identity_fn
    if reg_to_site is None:
        reg_to_site = _identity_fn

    terms_jordan_wigner = {}

    for term, coeff in terms.items():
        if not term:
            # all identity term, can't unzip
            terms_jordan_wigner[term] = coeff
            continue

        ops, _ = zip(*term)
        if {"+", "-"}.intersection(ops):
            # need to insert jordan-wigner strings
            new_term = []
            for op, site in term:
                reg = site_to_reg(site)
                if op in {"+", "-"}:
                    for r in range(reg):
                        site_below = reg_to_site(r)
                        new_term.append(("z", site_below))
                new_term.append((op, site))
            terms_jordan_wigner[tuple(new_term)] = coeff
        else:
            # no creation/annihilation operators, just add the term
            terms_jordan_wigner[term] = coeff

    return terms_jordan_wigner


@functools.lru_cache(maxsize=2**14)
def simplify_single_site_ops(coeff, ops):
    """Simplify a sequence of operators acting on the same site.

    Parameters
    ----------
    coeff : float or complex
        The coefficient of the operator sequence.
    ops : tuple of str
        The operator sequence.

    Returns
    -------
    new_coeff : float or complex
        The simplified coefficient.
    new_op : str
        The single, simplified operator that the sequence maps to, up to
        scalar multiplication.

    Examples
    --------

        >>> simplify_single_site_ops(1.0, ('+', 'z', 'z', 'z', 'z', '-'))
        (1.0, 'n')

        >>> simplify_single_site_ops(1.0, ("x", "y", "z"))
        (-1j, 'I')

    """

    if len(ops) == 1:
        return coeff, ops[0]

    # product all the matrices
    combo_mat = functools.reduce(operator.matmul, map(get_mat, ops))
    combo_coeff = combo_mat.flat[np.argmax(np.abs(combo_mat))]

    if combo_coeff == 0.0:
        # null-term
        return 0, None

    # find the reference operator that maps to this matrix
    for op in _OPMAP.keys():
        ref_mat = get_mat(op)
        ref_coeff = ref_mat.flat[np.argmax(np.abs(ref_mat))]
        if (
            (combo_mat / combo_coeff).round(12)
            == (ref_mat / ref_coeff).round(12)
        ).all():
            break
    else:
        raise ValueError(f"No match found for '{ops}'")

    coeff *= ref_coeff / combo_coeff
    if coeff.imag == 0.0:
        coeff = coeff.real

    return coeff, op


def simplify(terms, atol=1e-12, site_to_reg=None):
    """Simplify the given terms by combining operators acting on the same site,
    putting them in canonical order, removing null-terms, and combining
    equivalent operator strings.

    Parameters
    ----------
    terms : dict[term, coeff]
        The terms of the operator. Each term is a tuple of tuples, where each
        inner tuple is a pair of ``(operator, site)``.
    atol : float, optional
        The absolute tolerance for considering coefficients after
        simplification to be null.
    site_to_reg : callable, optional
        A function that maps a site to a linear register index. If not
        provided, the sites are assumed to be linear integers already.
        This is just used to sort the operators into canonical order.

    Returns
    -------
    terms_simplified : dict[term, coeff]
        The simplified terms of the operator. Each term is a tuple of tuples,
        where each inner tuple is a pair of ``(operator, site)``.
    """
    if site_to_reg is None:
        site_to_reg = _identity_fn

    terms_simplified = {}

    for term, coeff in terms.items():
        # collect operators acting on the same site
        collected = {}
        for op, site in term:
            collected.setdefault(site, []).append(op)

        # simplify operators acting on the same site & don't add null-terms
        simplified_ops = []
        for site, collected_ops in collected.items():
            coeff, op = simplify_single_site_ops(coeff, tuple(collected_ops))

            if op is None:
                # null-term ('e.g. '++' or '--')
                coeff = 0.0
                break

            if op != "I":
                # only need to record non-identity operators
                simplified_ops.append((op, site))

        if abs(coeff) < atol:
            # null-term
            continue

        # assume we can sort the operators into canonical order now
        simplified_ops.sort(key=lambda x: (site_to_reg(x[1]), x[0]))
        # combine coefficients of equivalent terms
        key = tuple(simplified_ops)
        coeff = terms_simplified.pop(key, 0.0) + coeff

        if abs(coeff) < atol:
            # null-term after combining coefficients
            continue

        if abs(coeff.imag) < atol:
            # if the coefficient is real, convert to real
            coeff = coeff.real

        terms_simplified[key] = coeff

    return terms_simplified


@functools.lru_cache(maxsize=None)
def get_pauli_decomp(op, atol=1e-12, use_zx=False):
    """Decompose the given operator (specified as a label) into a sum of
    Pauli components.

    Parameters
    ----------
    op : str
        The operator to decompose.
    atol : float, optional
        The absolute tolerance for considering coefficients to be null.
    use_zx : bool, optional
        Whether to decompose in terms of the real `ⴵ = ZX = iY` instead of `Y`.

    Returns
    -------
    list[tuple[float, str]]
        The decomposition of the operator into Pauli components. Each tuple
        contains the coefficient and the Pauli operator label.
    """
    bops = ("I", "x", "y", "z")

    if op in bops:
        terms = [(1.0, op)]
    else:
        terms = []
        mat = get_mat(op)
        for bop in bops:
            bmat = get_mat(bop)

            # Hilbert-Schmidt inner product
            cb = np.trace(bmat @ mat) / 2

            if abs(cb.imag) < atol:
                # realify if possible
                cb = cb.real

            # ignore zero components
            if abs(cb) >= atol:
                terms.append((cb, bop))

    if use_zx:
        # convert Y -> -iZX
        terms = [
            (-1j * coeff, "ⴵ") if op == "y" else (coeff, op)
            for coeff, op in terms
        ]

    return terms


def pauli_decompose(terms, atol=1e-12, use_zx=False, site_to_reg=None):
    """Decompose the given terms into a sum of Pauli strings.

    Parameters
    ----------
    terms : dict[term, coeff]
        The terms of the operator. Each term is a tuple of tuples, where each
        inner tuple is a pair of ``(operator, site)``.
    atol : float, optional
        The absolute tolerance for considering coefficients after
        decomposition to be null.
    use_xz : bool, optional
        Whether to decompose in terms of the real `ⴵ = ZX = iY` instead of `Y`.
    site_to_reg : callable, optional
        A function that maps a site to a linear register index. If not
        provided, the sites are assumed to be linear integers already.
        This is just used to sort the operators into canonical order.

    Returns
    -------
    terms_pauli_decomposed : dict[term, coeff]
        The decomposed terms of the operator. Each term is a tuple of tuples,
        where each inner tuple is a pair of ``(operator, site)``.
    """
    terms_pauli_decomposed = {}

    for ops, coeff in terms.items():
        # for each current term -> turn into potential sum
        new_ts = [(coeff, ())]

        for op, reg in ops:
            # for each operator in the string
            new_ts = [
                # extend with weighted pauli ...
                (coeff_t * dcoeff, (*ops_t, (dop, reg)))
                # ... for each weighted pauli in the decomposition
                for dcoeff, dop in get_pauli_decomp(op, atol, use_zx)
                # ... for each term in current sum
                for coeff_t, ops_t in new_ts
            ]

        for coeff, ops in new_ts:
            # we can sort paulis strings into canonical order, strip identities
            key = tuple(
                (op, site)
                for op, site in sorted(
                    ops, key=lambda x: (site_to_reg(x[1]), x[0])
                )
                if op != "I"
            )

            coeff = terms_pauli_decomposed.pop(key, 0.0) + coeff

            if abs(coeff) < atol:
                # null-term after combining coefficients
                continue

            if abs(coeff.imag) < atol:
                # if the coefficient is real, convert to real
                coeff = coeff.real

            terms_pauli_decomposed[key] = coeff

    return terms_pauli_decomposed


def get_pool_and_world_size(parallel):
    # figure out how many threads to use etc.
    from quimb import get_thread_pool

    if parallel is True:
        n_thread_workers = None
    elif isinstance(parallel, int):
        n_thread_workers = parallel
    else:
        raise ValueError(f"Unknown parallel option {parallel}.")

    pool = get_thread_pool(n_thread_workers)
    n_thread_workers = pool._max_workers

    return pool, n_thread_workers


def build_coupling_numba(term_store, site_to_reg, dtype=None):
    """Create a sparse nested dictionary of how each term couples each
    local site configuration to which other local site configuration, and
    with what coefficient, suitable for use with numba.

    Parameters
    ----------
    term_store : dict[term, coeff]
        The terms of the operator.
    site_to_reg : callable
        A function that maps a site to a linear register index.

    Returns
    -------
    coupling_map : tuple[ndarray]
        The operator defined as tuple of flat arrays.
    """
    if (dtype is None) or np.issubdtype(dtype, np.float64):
        dtype = np.float64
    elif np.issubdtype(dtype, np.complex128):
        dtype = np.complex128
    elif np.issubdtype(dtype, np.complex64):
        dtype = np.complex64
    elif np.issubdtype(dtype, np.float32):
        dtype = np.float32
    else:
        raise ValueError(f"Unknown dtype {dtype}")

    # number of operators per term e.g. 5 for '+zzz-'
    sizes_term = []
    # the registers each term acts on
    regs = []
    # number of elements per operator e.g. 2 for 'z', 1 for '+'
    sizes_op = []
    # input / output bits, e.g. 0 -> 1 for 'sx'
    xis = []
    xjs = []
    # transition coeffs, e.g. 0.5 for 'sx'
    cijs = []

    # for term t ...
    for term, coeff in term_store.items():
        if len(term) == 0:
            # special case: all identity term
            term = [("I", 0)]
            # directly add to first *register*
            map_to_reg = False
        else:
            map_to_reg = True

        # what sites does this term act non-trivially on?
        size = 0
        first_reg = True
        # which couples sites with product of ops ...
        for op, site in term:
            if map_to_reg:
                reg = site_to_reg(site)
            else:
                # special all identity term -> always first register
                reg = site

            regs.append(reg)
            # -> bit `xi` at `reg` is coupled to `xj` with coeff `cij`
            #          : reg
            #     ...10010...    xi=0  ->
            #     ...10110...    xj=1  with coeff cij

            # populate just the term/reg/bit maps we need
            size_t = 0
            for xi, (xj, cij) in _OPMAP[op].items():
                if first_reg:
                    # absorb overall coefficient into first coupling
                    cij = coeff * cij

                xis.append(xi)
                xjs.append(xj)
                cijs.append(cij)
                size_t += 1

            sizes_op.append(size_t)
            size += 1
            first_reg = False

        sizes_term.append(size)

    return (
        np.array(sizes_term, dtype=np.uint32),
        np.array(regs, dtype=np.uint32),
        np.array(sizes_op, dtype=np.uint8),
        np.array(xis, dtype=np.uint8),
        np.array(xjs, dtype=np.uint8),
        np.array(cijs, dtype=dtype),
    )


@functools.cache
def calc_dtype_cached(dtype, iscomplex):
    if dtype is None:
        # choose automatically
        if iscomplex:
            dtype = np.complex128
        else:
            dtype = np.float64
    else:
        if np.issubdtype(dtype, np.float64):
            dtype = np.float64
        elif np.issubdtype(dtype, np.complex128):
            dtype = np.complex128
        elif np.issubdtype(dtype, np.complex64):
            dtype = np.complex64
        elif np.issubdtype(dtype, np.float32):
            dtype = np.float32
        else:
            raise TypeError(f"Unsupported dtype {dtype}.")

    return dtype


class SparseOperatorBuilder:
    """Object for building operators with sparse structure. Specifically,
    a sum of terms, where each term is a product of operators, where each of
    these local operators acts on a single site and has at most one entry per
    row.

    Parameters
    ----------
    terms : sequence, optional
        The terms to initialize the builder with. ``add_term`` is simply called
        on each of these.
    hilbert_space : HilbertSpace
        The Hilbert space to build the operator in. If this is not supplied
        then a minimal Hilbert space will be constructed from the sites used,
        when required. The default symmetry and sector to build operators in is
        inherited from this Hilbert space, but can be overridden.
    dtype : numpy.dtype or str, optional
        A default data type for the coefficients of the operator. If not
        provided, will be automatically determined at building time based on
        the terms in the operator. If the operator is complex, will be set to
        ``np.complex128``. If the operator is real, will be set to
        ``np.float64``. Individual building methods can override this.
    jordan_wigner : bool, optional
        Whether to apply the Jordan-Wigner transformation to the terms
        automatically when processing them. This prepends pauli Z strings to
        all creation and annihilation operators.
    pauli_decompose : bool or "zx", optional
        Whether to apply the Pauli decomposition to the terms automatically
        when processing them. This decomposes all local operators into sums of
        Pauli operators. If "zx" is supplied, the decomposition is done in
        terms of the real `ZX = iY` operator instead of `Y`.
    atol : float, optional
        The absolute tolerance for considering coefficients to be null when
        simplifying and decomposing terms.
    """

    def __init__(
        self,
        terms=(),
        hilbert_space: HilbertSpace = None,
        dtype=None,
        jordan_wigner=False,
        pauli_decompose=False,
        atol=1e-12,
    ):
        self._sites_used = set()
        self._hilbert_space = hilbert_space

        # terms as they are supplied by user
        self._terms_raw = {}
        # terms after processing (jordan-wigner, pauli decomp, simplification)
        self._terms_final = None

        # processing flags
        self._transform_jordan_wigner = jordan_wigner
        self._transform_pauli_decompose = pauli_decompose
        self._atol = atol

        self._dtype = dtype
        self._coupling_maps = {}
        self._cache = {}

        for term in terms:
            self.add_term(*term)

    @property
    def sites_used(self):
        """A tuple of the sorted coordinates/sites seen so far."""
        return tuple(sorted(self._sites_used))

    @property
    def hilbert_space(self) -> HilbertSpace:
        """The Hilbert space of the operator. Created from the sites seen
        so far if not supplied at construction.
        """
        if self._hilbert_space is None:
            self._hilbert_space = HilbertSpace(self.sites_used)
        return self._hilbert_space

    @property
    def nsites(self):
        """The total number of coordinates/sites seen so far."""
        return self.hilbert_space.nsites

    def site_to_reg(self, site):
        """Get the register / linear index of coordinate ``site``."""
        return self.hilbert_space.site_to_reg(site)

    def reg_to_site(self, reg):
        """Get the site of register / linear index ``reg``."""
        return self.hilbert_space.reg_to_site(reg)

    @property
    def terms_raw(self):
        """A tuple of the raw terms seen so far, as a mapping from
        operator strings to coefficients.
        """
        return tuple((coeff, ops) for ops, coeff in self._terms_raw.items())

    def _reset_caches(self):
        """Reset any cached representations of the operator, used whenever the
        terms are modified in any way, and thus require reprocessing.
        """
        self._cache.clear()
        self._coupling_maps.clear()
        self._terms_final = None

    def _get_terms_final(self):
        """Get the processed terms, applying any requested transformations, if
        not already done.
        """
        if self._terms_final is None:
            # need to (re)process raw terms to final terms
            terms = self._terms_raw

            # 1. jordan wigner transform before anything else
            if self._transform_jordan_wigner:
                terms = jordan_wigner_transform(
                    terms=terms,
                    site_to_reg=self.site_to_reg,
                    reg_to_site=self.reg_to_site,
                )

            terms = simplify(
                terms=terms,
                atol=self._atol,
                site_to_reg=self.site_to_reg,
            )

            # 2. pauli decomposition next
            if self._transform_pauli_decompose:
                terms = pauli_decompose(
                    terms=terms,
                    atol=self._atol,
                    use_zx=self._transform_pauli_decompose == "zx",
                    site_to_reg=self.site_to_reg,
                )

            # 3. finally simplify into strict form last
            self._terms_final = simplify(
                terms=terms,
                atol=self._atol,
                site_to_reg=self.site_to_reg,
            )

        return self._terms_final

    @property
    def terms(self):
        """A tuple of the, possibly transformed, terms seen so far."""
        return tuple(
            (coeff, ops) for ops, coeff in self._get_terms_final().items()
        )

    @property
    def nterms(self):
        """The total number of terms seen so far."""
        return len(self._get_terms_final())

    @property
    def locality(self):
        """The locality of the operator, the maximum support of any term."""
        terms = self._get_terms_final()
        if not terms:
            return 0
        return max(len(ops) for ops in terms)

    @property
    def iscomplex(self):
        """Whether the operator has complex terms."""
        try:
            iscomplex = self._cache["iscomplex"]
        except KeyError:
            iscomplex = self._cache["iscomplex"] = any(
                np.iscomplexobj(coeff)
                or any(op in _OPCOMPLEX for op, _ in ops)
                for ops, coeff in self._get_terms_final().items()
            )
        return iscomplex

    def add_term(self, *coeff_ops, atol=1e-12):
        """Add a term to the operator.

        Parameters
        ----------
        coeff : float, optional
            The overall coefficient of the term.
        ops : sequence of tuple[str, hashable]
            The operators of the term, together with the sites they act on.
            Each term should be a pair of ``(operator, site)``, where
            ``operator`` can be:

            - ``'x'``, ``'y'``, ``'z'``: Pauli matrices
            - ``'sx'``, ``'sy'``, ``'sz'``: spin operators (i.e. scaled
                Pauli matrices)
            - ``'+'``, ``'-'``: creation/annihilation operators
            - ``'n'``, ``'sn'``, or ``'h'``: number, symmetric
                number (n - 1/2) and hole (1 - n) operators

            And ``site`` is a hashable object that represents the site that
            the operator acts on.

        """
        if isinstance(coeff_ops[0], (tuple, list)):
            # assume coeff is 1.0
            coeff = 1
            ops = coeff_ops
        else:
            coeff, *ops = coeff_ops
            if abs(coeff) < atol:
                # null-term
                return

        # parse the operator specification
        ops = tuple(tuple(op) for op in ops)
        for op, site in ops:
            # check that the site is valid if the Hilbert space is known
            if (
                self._hilbert_space is not None
            ) and not self._hilbert_space.has_site(site):
                raise ValueError(f"Site {site} not in the Hilbert space.")

            # record used sites, for later construction of the Hilbert space
            self._sites_used.add(site)

            if op not in _OPMAP:
                raise ValueError(f"Unknown operator '{op}'.")

        # if we have already seen this exact term, just add the coeff but note,
        # we do not simplify equivalent terms here, in case its a fermionic
        coeff = self._terms_raw.pop(ops, 0.0) + coeff

        if abs(coeff) < atol:
            # null-term after combining coefficients
            return

        if abs(coeff.imag) < atol:
            # if the coefficient is real, convert to real
            coeff = coeff.real

        self._terms_raw[ops] = coeff
        self._reset_caches()

    def __iadd__(self, term):
        self.add_term(*term)
        return self

    def __isub__(self, term):
        self.add_term(-term[0], *term[1:])
        return self

    def jordan_wigner_transform(self, value=None):
        """Toggle transforming the terms in this operator by pre-prending
        pauli Z strings to all creation and annihilation operators. This is
        always applied directly as the first processing step to the raw terms,
        so that the fermionic ordering is respected.

        Note this doesn't decompose +, - into (X + iY) / 2 and (X - iY) / 2, it
        just prepends Z strings. Use `pauli_decompose` to get the full
        decomposition.

        The placement of the Z strings is defined by the ordering of the
        hilbert space, by default, the sorted order of the site labels.

        Parameters
        ----------
        value : bool, optional
            Whether to apply the Jordan-Wigner transformation. If `None` (the
            default) then this method acts as toggle. Whereas supplying `True`
            or `False` explicitly sets or unsets the transformation.
        """
        if value is None:
            value = not self._transform_jordan_wigner
        else:
            value = bool(value)
        self._transform_jordan_wigner = value
        self._reset_caches()

    def pauli_decompose(self, value=None, atol=None, use_zx=False):
        """Transform the terms in this operator by decomposing them into
        Pauli strings.

        Parameters
        ----------
        value : bool, optional
            Whether to apply the Pauli decomposition. If `None` (the default)
            then this method acts as toggle. Whereas supplying `True` or
            `False` explicitly sets or unsets the transformation.
        """
        if value is None:
            value = not self._transform_pauli_decompose
        else:
            value = bool(value)

        if value and use_zx:
            self._transform_pauli_decompose = "zx"
        else:
            self._transform_pauli_decompose = value

        if atol is not None:
            self._atol = atol

        self._reset_caches()

    def show(self, filler="."):
        """Print an ascii representation of the terms in this operator."""
        print(self)
        for _, (term, coeff) in enumerate(self._get_terms_final().items()):
            s = [f"{filler} "] * self.nsites
            for op, site in term:
                s[self.site_to_reg(site)] = f"{op:<2}"
            print("".join(s), f"{coeff:+}")

    def get_dtype(self, dtype=None):
        """Calculate the numpy data type of the operator to use.

        Parameters
        ----------
        dtype : numpy.dtype or str, optional
            The data type of the coefficients. If not provided, will be
            automatically determined based on the terms in the operator.

        Returns
        -------
        dtype : numpy.dtype
            The data type of the coefficients.
        """
        if dtype is None:
            dtype = self._dtype
        return calc_dtype_cached(dtype, self.iscomplex)

    def get_coupling_map(self, dtype=None):
        """Build and cache the coupling map for the specified dtype.

        Parameters
        ----------
        dtype : numpy.dtype, optional
            The data type of the coefficients. If not provided, will be
            automatically determined based on the terms in the operator.

        Returns
        -------
        coupling_map : tuple[ndarray]
            The operator defined as tuple of flat arrays.
        """
        dtype = self.get_dtype(dtype)

        try:
            coupling_map = self._coupling_maps[dtype]
        except KeyError:
            coupling_map = build_coupling_numba(
                self._get_terms_final(),
                self.site_to_reg,
                dtype=dtype,
            )
            self._coupling_maps[dtype] = coupling_map

        return coupling_map

    def flatconfig_coupling(self, flatconfig, dtype=None):
        """Get an array of other configurations coupled to the given individual
        ``flatconfig`` by this operator, and the corresponding coupling
        coefficients. This is for use with VMC for example.

        Parameters
        ----------
        flatconfig : array[np.uint8]
            The linear array of the configuration to get the coupling for.
        dtype : numpy.dtype, optional
            The data type of the matrix. If not provided, will be
            automatically determined based on the terms in the operator.

        Returns
        -------
        coupled_flatconfigs : ndarray[np.uint8]
            Each distinct flatconfig coupled to ``flatconfig``.
        coeffs : ndarray[dtype]
            The corresponding coupling coefficients.
        """
        dtype = self.get_dtype(dtype)
        coupling_map = self.get_coupling_map(dtype=dtype)
        return configcore.flatconfig_coupling_numba(
            flatconfig,
            coupling_map=coupling_map,
            dtype=dtype,
        )

    def config_coupling(self, config, dtype=None):
        """Get a list of other configurations coupled to ``config`` by this
        operator, and the corresponding coupling coefficients. This is for
        use with VMC for example.

        Parameters
        ----------
        config : dict[site, int]
            The configuration to get the coupling for.

        Returns
        -------
        coupled_configs : list[dict[site, np.uint8]]
            Each distinct configuration coupled to ``config``.
        coeffs: list[dtype]
            The corresponding coupling coefficients.
        """
        fc = self.hilbert_space.config_to_flatconfig(config)
        bjs, cijs = self.flatconfig_coupling(fc, dtype=dtype)
        coupled_configs = [
            self.hilbert_space.flatconfig_to_config(bj) for bj in bjs
        ]
        return coupled_configs, cijs

    def evaluate_exact_flatconfigs(self, fn_amplitude, progbar=False):
        """Calculate the expectation value of this operator with respect to
        a wavefunction provided as a function with signature::

            fn_amplitude(flatconfig: ndarray[np.uint8]) -> z: float | complex

        """
        O = 0.0
        p = 0.0

        if progbar:
            import tqdm

            iterator = tqdm.tqdm(range(self.hilbert_space.size))
        else:
            iterator = range(self.hilbert_space.size)

        for r in iterator:
            flatconfig = self.hilbert_space.rank_to_flatconfig(r)

            xpsi = fn_amplitude(flatconfig)
            if not xpsi:
                continue

            pi = abs(xpsi) ** 2
            p += pi

            Oloc = 0.0
            for fy, hxy in zip(*self.flatconfig_coupling(flatconfig)):
                ypsi = fn_amplitude(fy)
                Oloc = Oloc + hxy * ypsi / xpsi

            O += Oloc * pi

        return O / p

    def evaluate_exact_configs(self, fn_amplitude, progbar=False):
        """Calculate the expectation value of this operator with respect to
        a wavefunction provided as a function with signature::

            fn_amplitude(config: dict[site, int]) -> z: float | complex

        """
        O = 0.0
        p = 0.0

        if progbar:
            import tqdm

            iterator = tqdm.tqdm(range(self.hilbert_space.size))
        else:
            iterator = range(self.hilbert_space.size)

        for r in iterator:
            config = self.hilbert_space.rank_to_config(r)

            xpsi = fn_amplitude(config)
            if not xpsi:
                continue

            pi = abs(xpsi) ** 2
            p += pi

            Oloc = 0.0
            for fy, hxy in zip(*self.config_coupling(config)):
                ypsi = fn_amplitude(fy)
                Oloc = Oloc + hxy * ypsi / xpsi

            O += Oloc * pi

        return O / p

    def build_coo_data(
        self, sector=None, symmetry=None, dtype=None, parallel=False
    ):
        """Build the raw data for a sparse matrix in COO format, optionally
        in parallel.

        Parameters
        ----------
        sector : {None, str, int, ((int, int), (int, int))}, optional
            The sector of the Hilbert space. If None, the default sector is
            used.
        symmetry : {None, "Z2", "U1", "U1U1"}, optional
            The symmetry of the Hilbert space. If None, the default symmetry is
            used, or inferred from the supplied sector if possible.
        dtype : numpy.dtype, optional
            The data type of the matrix. If not provided, will be
            automatically determined based on the terms in the operator.
        parallel : bool or int, optional
            Whether to build the matrix in parallel (multi-threaded). If True,
            will use number of threads equal to the number of available CPU
            cores. If False, will use a single thread. If an integer is
            provided, it will be used as the number of threads to use.

        Returns
        -------
        data : array
            The data entries for the sparse matrix in COO format.
        rows : array
            The row indices for the sparse matrix in COO format.
        cols : array
            The column indices for the sparse matrix in COO format.
        d : int
            The total number of basis states.
        """
        dtype = self.get_dtype(dtype)
        coupling_map = self.get_coupling_map(dtype=dtype)
        d = self.hilbert_space.get_size(sector, symmetry)
        sector_nb, symmetry_nb = self.hilbert_space.get_sector_numba(
            sector=sector, symmetry=symmetry
        )
        kwargs = {
            "coupling_map": coupling_map,
            "sector": sector_nb,
            "symmetry": symmetry_nb,
            "dtype": dtype,
        }

        if not parallel:
            data, rows, cols = configcore.build_coo_numba_core(**kwargs)
        else:
            pool, world_size = get_pool_and_world_size(parallel)

            # launch the threads! note we distribtue in cyclic fashion as the
            # sparsity can be concentrated in certain ranges and we want each
            # thread to have roughly the same amount of work to do
            fs = [
                pool.submit(
                    configcore.build_coo_numba_core,
                    world_rank=i,
                    world_size=world_size,
                    **kwargs,
                )
                for i in range(world_size)
            ]

            # gather and concatenate the results (some memory overhead here)
            data = []
            rows = []
            cols = []
            for f in fs:
                df, ci, cj = f.result()
                data.append(df)
                rows.append(ci)
                cols.append(cj)

            data = np.concatenate(data)
            rows = np.concatenate(rows)
            cols = np.concatenate(cols)

        return data, rows, cols, d

    def build_sparse_matrix(
        self,
        sector=None,
        symmetry=None,
        dtype=None,
        stype="csr",
        parallel=False,
    ):
        """Build a sparse matrix in the given format. Optionally in parallel.

        Parameters
        ----------
        sector : {None, str, int, ((int, int), (int, int))}, optional
            The sector of the Hilbert space. If None, the default sector is
            used.
        symmetry : {None, "Z2", "U1", "U1U1"}, optional
            The symmetry of the Hilbert space. If None, the default symmetry is
            used, or inferred from the supplied sector if possible.
        dtype : numpy.dtype, optional
            The data type of the matrix. If not provided, will be
            automatically determined based on the terms in the operator.
        stype : str, optional
            The sparse matrix format to use. Can be one of 'coo', 'csr', 'csc',
            'bsr', 'lil', 'dok', or 'dia'. Default is 'csr'.
        parallel : bool, optional
            Whether to build the matrix in parallel (multi-threaded).

        Returns
        -------
        scipy.sparse matrix
        """
        import scipy.sparse as sp

        data, rows, cols, d = self.build_coo_data(
            sector=sector,
            symmetry=symmetry,
            dtype=dtype,
            parallel=parallel,
        )
        A = sp.coo_matrix((data, (rows, cols)), shape=(d, d))
        if stype != "coo":
            A = A.asformat(stype)

        return A

    def build_dense(
        self,
        sector=None,
        symmetry=None,
        dtype=None,
        parallel=False,
    ):
        """Get the dense (`numpy.ndarray`) matrix representation of this
        operator.

        Parameters
        ----------
        sector : {None, str, int, ((int, int), (int, int))}, optional
            The sector of the Hilbert space. If None, the default sector is
            used.
        symmetry : {None, "Z2", "U1", "U1U1"}, optional
            The symmetry of the Hilbert space. If None, the default symmetry is
            used, or inferred from the supplied sector if possible.
        dtype : numpy.dtype, optional
            The data type of the matrix. If not provided, will be
            automatically determined based on the terms in the operator.
        parallel : bool or int, optional
            Whether to build the matrix in parallel (multi-threaded). If
            True, will use number of threads equal to the number of
            available CPU cores. If False, will use a single thread. If an
            integer is provided, it will be used as the number of threads to
            use.

        Returns
        -------
        A : numpy.ndarray
            The dense matrix representation of this operator.
        """
        A = self.build_sparse_matrix(
            sector=sector,
            symmetry=symmetry,
            dtype=dtype,
            stype="coo",
            parallel=parallel,
        )
        return A.toarray()

    def matvec(
        self,
        x,
        out=None,
        sector=None,
        symmetry=None,
        dtype=None,
        parallel=False,
    ):
        """Apply this operator lazily (i.e. without constructing a sparse
        matrix) to a vector. This uses less memory but is much slower.

        Parameters
        ----------
        x : array
            The vector to apply the operator to.
        out : array, optional
            An array to store the result in. If not provided, a new array
            will be created.
        sector : {None, str, int, ((int, int), (int, int))}, optional
            The sector of the Hilbert space. If None, the default sector is
            used. The implicit size should match that of `x`.
        symmetry : {None, "Z2", "U1", "U1U1"}, optional
            The symmetry of the Hilbert space. If None, the default symmetry is
            used, or inferred from the supplied sector if possible.
        dtype : numpy.dtype, optional
            The data type of the matrix. If not provided, will be
            automatically set as the same as the input vector.
        parallel : bool or int, optional
            Whether to apply the operator in parallel (multi-threaded). If
            True, will use number of threads equal to the number of
            available CPU cores. If False, will use a single thread. If an
            integer is provided, it will be used as the number of threads to
            use. Uses `num_threads` more memory but is faster.

        Returns
        -------
        out : array
            The result of applying the operator to the vector.
        """
        if dtype is None:
            dtype = x.dtype

        sector_nb, symmetry_nb = self.hilbert_space.get_sector_numba(
            sector=sector, symmetry=symmetry
        )

        kwargs = {
            "coupling_map": self.get_coupling_map(dtype=dtype),
            "sector": sector_nb,
            "symmetry": symmetry_nb,
        }

        if not parallel:
            if out is None:
                out = np.zeros_like(x, dtype=dtype)
            configcore.matvec_numba(x, out, **kwargs)
            return out

        pool, world_size = get_pool_and_world_size(parallel)
        out_i = np.zeros_like(x, dtype=dtype, shape=(world_size, x.size))

        fs = [
            pool.submit(
                configcore.matvec_numba,
                x,
                out_i[i],
                world_rank=i,
                world_size=world_size,
                **kwargs,
            )
            for i in range(world_size)
        ]
        for f in fs:
            f.result()

        # sum the results from each thread
        return np.sum(out_i, axis=0, out=out)

    def aslinearoperator(
        self,
        sector=None,
        symmetry=None,
        dtype=None,
        parallel=False,
    ):
        """Get a `scipy.sparse.linalg.LinearOperator` for this operator.
        This is a lazy representation of the operator, which uses `matvec` to
        apply the operator to a vector. Less memory is required than
        constructing the full sparse matrix, but it is significantly slower.

        Note currently the operator is assumed to be hermitian.

        Parameters
        ----------
        sector : {None, str, int, ((int, int), (int, int))}, optional
            The sector of the Hilbert space. If None, the default sector is
            used.
        symmetry : {None, "Z2", "U1", "U1U1"}, optional
            The symmetry of the Hilbert space. If None, the default symmetry is
            used, or inferred from the supplied sector if possible.
        dtype : numpy.dtype, optional
            The data type of the matrix. If not provided, will be
            automatically determined based on the terms in the operator.
        parallel : bool or int, optional
            Whether to apply the operator in parallel (multi-threaded). If
            True, will use number of threads equal to the number of
            available CPU cores. If False, will use a single thread. If an
            integer is provided, it will be used as the number of threads to
            use. Uses `num_threads` more memory but is faster.

        Returns
        -------
        Alo : scipy.sparse.linalg.LinearOperator
            The linear operator representation of this operator.
        """
        from scipy.sparse.linalg import LinearOperator

        if dtype is None:
            dtype = self.get_dtype()

        d = self.hilbert_space.get_size(sector, symmetry)
        shape = (d, d)

        matvec = functools.partial(
            self.matvec,
            sector=sector,
            symmetry=symmetry,
            dtype=dtype,
            parallel=parallel,
        )

        return LinearOperator(
            shape=shape,
            matvec=matvec,
            rmatvec=matvec,
            dtype=dtype,
        )

    def build_local_terms(self, dtype=None):
        """Get a dictionary of local terms, where each key is a sorted tuple
        of sites, and each value is the local matrix representation of the
        operator on those sites. For use with e.g. tensor network algorithms.

        Note terms acting on the same sites are summed together and the size of
        each local matrix is exponential in the locality of that term.

        Returns
        -------
        Hk : dict[tuple[hashable], numpy.ndarray]
            The local terms.
        """
        Hk = {}

        dtype = self.get_dtype(dtype)

        for term, coeff in self._get_terms_final().items():
            ops, sites = zip(*term)
            mats = (get_mat(op, dtype=dtype) for op in ops)
            hk = coeff * functools.reduce(np.kron, mats)
            if sites not in Hk:
                Hk[sites] = hk
            else:
                Hk[sites] = Hk[sites] + hk

        return Hk

    def build_local_ham(self, dtype=None):
        """Get a `LocalHamGen` object for this operator.

        Parameters
        ----------
        dtype : numpy.dtype, optional
            The data type of the matrix. If not provided, will be
            automatically determined based on the terms in the operator.

        Returns
        -------
        H : LocalHamGen
            The local Hamiltonian representation of this operator.
        """
        from quimb.tensor.tensor_arbgeom_tebd import LocalHamGen

        terms = self.build_local_terms(dtype=dtype)

        H2 = {}
        H1 = {}
        for sites, hk in terms.items():
            if len(sites) == 2:
                H2[sites] = hk
            elif len(sites) == 1:
                H1[sites[0]] = hk
            else:
                raise NotImplementedError(
                    "Only supports 1- and 2-site terms for now."
                )

        if not H1:
            H1 = None
        if not H2:
            H2 = None

        return LocalHamGen(H2, H1)

    def build_state_machine_greedy(self, atol=1e-12):
        # XXX: also implement optimal method : https://arxiv.org/abs/2006.02056

        import networkx as nx

        # - nodes of the state machine are a 2D grid of (register, 'rail'),
        #   with the maximum number of rails giving the eventual bond dimension
        # - there are N + 1 registers for N sites
        # - each edge from (r, i) to (r + 1, j) represents a term that will be
        #   placed in the rth MPO tensor at entry like W[i, j, :, :] = op
        # - each node has either a single inwards or outwards edge
        G = nx.DiGraph()
        G.add_node((0, 0))

        # count how many rails are at each register
        num_rails = [1] + [0] * self.nsites

        # track which terms pass through each edge and vice versa
        edges_to_terms = {}
        terms_to_edges = {}
        # so that we can place all the coefficients at the end
        coeffs_to_place = {}

        def new_edge(a, b):
            # need to track which terms pass through this edge so we can
            # place the coefficient somewhere unique at the end
            G.add_edge(a, b, op=op, weight=1, coeff=None)
            edges_to_terms.setdefault((a, b), set()).add(t)
            terms_to_edges.setdefault(t, set()).add((a, b))

            if G.out_degree(a) > 1:
                # no longer valid out_string, for all ancestor nodes
                G.nodes[a]["out_string"] = None
                for prev_node in nx.ancestors(G, a):
                    G.nodes[prev_node]["out_string"] = None

        def check_right():
            # check if can **right share**
            # - check all existing potential next nodes
            # - right strings must match
            # - must be single output node
            # - current op must match or not exist
            for rail in range(num_rails[reg + 1]):
                cand_node = (reg + 1, rail)
                if G.out_degree(cand_node) > 1:
                    # can't share if there's not a single output string
                    continue

                if G.nodes[cand_node]["out_string"] != string[reg + 1 :]:
                    # output string must match
                    continue

                e = (current_node, cand_node)
                if e not in G.edges:
                    new_edge(current_node, cand_node)
                    return cand_node
                else:
                    if G.edges[e]["op"] != op:
                        continue
                    G.edges[e]["weight"] += 1
                    edges_to_terms.setdefault(e, set()).add(t)
                    terms_to_edges.setdefault(t, set()).add(e)
                    return cand_node

                # XXX: if we can right share, don't need to do anything
                # more since whole remaining string is shared?
                # -> possibly, to track term congestion for coeff placement

        def check_left():
            # check if can **left share**
            # - check all out edges
            # - current op must match AND
            # - must be single input node
            for e in G.edges(current_node):
                cand_node = e[1]
                if G.in_degree(cand_node) <= 1:
                    if G.edges[e]["op"] == op:
                        G.edges[e]["weight"] += 1
                        edges_to_terms.setdefault(e, set()).add(t)
                        terms_to_edges.setdefault(t, set()).add(e)
                        return cand_node

        def create_new():
            # create a new rail at the next register
            next_node = (reg + 1, num_rails[reg + 1])
            num_rails[reg + 1] += 1
            new_edge(current_node, next_node)

            # if G.out_degree(current_node) > 1:
            #     # no longer valid out_string, for all ancestor nodes
            #     G.nodes[current_node]["out_string"] = None
            #     for prev_node in nx.ancestors(G, current_node):
            #         G.nodes[prev_node]["out_string"] = None

            # the new node is always a single output node so far
            G.nodes[next_node]["out_string"] = string[reg + 1 :]

            return next_node

        for t, (term, coeff) in enumerate(self._get_terms_final().items()):
            # build full string for this term including identity ops
            rmap = {self.site_to_reg(site): op for op, site in term}
            string = tuple(rmap.get(r, "I") for r in range(self.nsites))

            current_node = (0, 0)
            for reg, op in enumerate(string):
                cand_node = check_right()
                if cand_node is not None:
                    # can share right part of string
                    current_node = cand_node
                else:
                    cand_node = check_left()
                    if cand_node is not None:
                        # can share left part of string
                        current_node = cand_node
                    else:
                        # have to create new node
                        current_node = create_new()

            if coeff != 1.0:
                # record that we still need to place coeff somewhere
                coeffs_to_place[t] = coeff

        G.graph["nsites"] = self.nsites
        G.graph["num_rails"] = tuple(num_rails)
        G.graph["max_num_rails"] = max(num_rails)

        # how many terms pass through each edge
        edge_scores = {e: len(ts) for e, ts in edges_to_terms.items()}

        # the least congested edge a term passes through
        term_scores = {
            t: min(edge_scores[e] for e in es)
            for t, es in terms_to_edges.items()
        }

        def place_coeff(edge, coeff):
            G.edges[edge]["coeff"] = coeff

            # every term passing through this edge is multiplied by this coeff
            for t in edges_to_terms.pop(edge):
                new_coeff = coeffs_to_place.pop(t, 1.0) / coeff
                if abs(new_coeff - 1.0) > atol:
                    # if another term doesn't have matching coeff, still need
                    # to place the updated coeff
                    coeffs_to_place[t] = new_coeff

                # # remove edge as candidate for placing other coefficients
                # terms_to_edges[t].remove(edge)
            edge_scores.pop(edge)

        while coeffs_to_place:
            # get the remaining term with the maximum congestion
            t = max(coeffs_to_place, key=term_scores.get)
            # get the least congested edge it passes through
            best = min(
                (e for e in terms_to_edges[t] if e in edge_scores),
                key=edge_scores.get,
            )
            # place it and update everything
            place_coeff(best, coeffs_to_place[t])

        return G

    def draw_state_machine(
        self,
        method="greedy",
        figsize="auto",
        G=None,
    ):
        """Draw the fintie state machine for this operator, as if buildling
        the MPO.
        """
        import math

        from matplotlib import pyplot as plt

        from quimb.schematic import auto_colors

        if G is None:
            if method == "greedy":
                G = self.build_state_machine_greedy()
            else:
                raise ValueError(f"Unknown method {method}")

        def labelled_arrow(ax, p1, p2, label, color, width):
            angle = math.atan2(p2[1] - p1[1], p2[0] - p1[0])
            ax.annotate(
                "",
                xy=p2,
                xycoords="data",
                xytext=p1,
                textcoords="data",
                arrowprops=dict(
                    arrowstyle="->",
                    color=color,
                    alpha=0.75,
                    linewidth=width,
                ),
            )
            p_middle = (p1[0] + p2[0]) / 2, (p1[1] + p2[1]) / 2
            ax.text(
                *p_middle,
                label,
                color=color,
                ha="center",
                va="center",
                rotation=angle * 180 / math.pi,
                alpha=1.0,
                transform_rotates_text=True,
            )

        if figsize == "auto":
            width = G.graph["nsites"]
            # maximum number of nodes on any rail
            height = G.graph["max_num_rails"] / 2
            figsize = (width, height)

        fig, ax = plt.subplots(figsize=figsize)
        fig.patch.set_alpha(0.0)
        ax.set_axis_off()
        ax.set_xlim(-0.5, G.graph["nsites"] + 0.5)
        ax.set_ylim(-0.5, G.graph["max_num_rails"] + 0.5)

        # draw each edge as a colored, labelled node
        ops = sorted(
            {data["op"] for _, _, data in G.edges(data=True)} - {"I"}, key=str
        )
        all_colors = auto_colors(len(ops))
        colors = {op: c for op, c in zip(ops, all_colors)}
        colors["I"] = "grey"
        for n1, n2, data in G.edges(data=True):
            color = colors[data["op"]]
            width = math.log2(1 + data["weight"])
            label = data["op"]
            if data.get("coeff", None) is not None:
                label += f" * {data['coeff']:.4g}"
            label += "\n"
            labelled_arrow(ax, n1, n2, label, color, width)

        # label which MPO site along the bottom
        for i in range(G.graph["nsites"]):
            ax.text(
                i + 0.5,
                -1.0,
                "$W_{" + str(i) + "}$",
                ha="center",
                va="center",
                color=(0.5, 0.5, 0.5),
                fontsize=12,
            )

        plt.show()
        plt.close(fig)

    def build_mpo(self, method="greedy", dtype=None, **mpo_opts):
        """Build a matrix product operator (MPO) representation of this
        operator.

        Parameters
        ----------
        method : str, optional
            The method to use for building the MPO. Currently only "greedy"
            is supported.
        dtype : type, optional
            The data type of the MPO. If not supplied, will be chosen
            automatically based on the terms in the operator.
        mpo_opts : keyword arguments
            Additional options to pass to the MPO constructor.
            See `MatrixProductOperator` for details.

        Returns
        -------
        mpo : MatrixProductOperator
            The MPO representation of this operator.
        """
        import numpy as np

        import quimb as qu
        import quimb.tensor as qtn

        dtype = self.get_dtype(dtype)

        if self.nsites == 1:
            # single site operator, just sum them
            Wt0 = self.build_dense(dtype=dtype)
            return qtn.MatrixProductOperator([Wt0], **mpo_opts)

        if method == "greedy":
            G = self.build_state_machine_greedy()
        else:
            raise ValueError(f"Unknown method {method}.")

        Wts = [
            np.zeros((dl, dr, 2, 2), dtype=dtype)
            for dl, dr in qu.utils.pairwise(G.graph["num_rails"])
        ]

        for node_a, node_b, data in G.edges(data=True):
            op = data["op"]
            coeff = data.get("coeff", None)
            if coeff is not None:
                mat = coeff * get_mat(op)
            else:
                mat = get_mat(op)

            rega, raila = node_a
            _, railb = node_b
            Wts[rega][raila, railb, :, :] = mat

        Wts[0] = Wts[0].sum(axis=0)
        Wts[-1] = Wts[-1].sum(axis=1)

        return qtn.MatrixProductOperator(Wts, **mpo_opts)

    def __repr__(self):
        s = [f"{self.__class__.__name__}("]
        s.append(f"nsites={self.nsites}")
        s.append(f", nterms={self.nterms}")
        s.append(f", locality={self.locality}")

        if self._transform_jordan_wigner:
            s.append(", jordan_wigner=True")
        if self._transform_pauli_decompose:
            s.append(f", pauli_decompose={self._transform_pauli_decompose}")

        s.append(")")
        return "".join(s)

    def build_matrix_ikron(self, **ikron_opts):
        """Build either the dense or sparse matrix of this operator via
        explicit calls to `ikron`. This is a slower but useful alternative
        testing method.
        """
        from quimb import ikron

        A = None

        dims = [2] * self.nsites
        for ops, coeff in self._get_terms_final().items():
            if not ops:
                ops = [("I", 0)]

            mats = []
            inds = []
            for op, reg in ops:
                mats.append(get_mat(op))
                inds.append(reg)

            Aterm = coeff * ikron(mats, dims, inds, **ikron_opts)

            if A is None:
                A = Aterm
            else:
                A = A + Aterm

        return A



"""Some functions for generating the edges of a lattice."""

import itertools
import random


def sort_unique(edges):
    """Make sure there are no duplicate edges and that for each
    ``coo_a < coo_b``.
    """
    return tuple(
        sorted(tuple(sorted(edge)) for edge in set(map(frozenset, edges)))
    )


# ----------------------------------- 1D ------------------------------------ #


def edges_1d_chain(L, cyclic=False):
    """Return the graph edges of a finite 1D chain lattice.

    Parameters
    ----------
    L : int
        The number of cells.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.

    Returns
    -------
    edges : list[(int, int)]
    """
    edges = []
    for i in range(L):
        if i < L - 1 or cyclic:
            edges.append((i, (i + 1) % L))

    return sort_unique(edges)


# ----------------------------------- 2D ------------------------------------ #


def check_2d(coo, Lx, Ly, cyclic):
    """Check ``coo`` in inbounds for a maybe cyclic 2D lattice."""
    x, y = coo
    if (not cyclic) and not ((0 <= x < Lx) and (0 <= y < Ly)):
        return
    return (x % Lx, y % Ly)


def edges_2d_square(Lx, Ly, cyclic=False, cells=None):
    """Return the graph edges of a finite 2D square lattice. The nodes
    (sites) are labelled like ``(i, j)``.

    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly))``.

    Returns
    -------
    edges : list[((int, int), (int, int))]
    """
    if cells is None:
        cells = itertools.product(range(Lx), range(Ly))

    edges = []
    for i, j in cells:
        for coob in [(i, j + 1), (i + 1, j)]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j), coob))

    return sort_unique(edges)


def edges_2d_hexagonal(Lx, Ly, cyclic=False, cells=None):
    """Return the graph edges of a finite 2D hexagonal lattice. There are two
    sites per cell, and note the cells do not form a square tiling. The nodes
    (sites) are labelled like ``(i, j, s)`` for ``s`` in ``'AB'``.

    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly))``.

    Returns
    -------
    edges : list[((int, int, str), (int, int, str))]
    """
    if cells is None:
        cells = itertools.product(range(Lx), range(Ly))

    edges = []
    for i, j in cells:
        for *coob, lbl in [
            (i, j, "B"),
            (i, j - 1, "B"),
            (i - 1, j, "B"),
        ]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j, "A"), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, "A"),
            (i, j + 1, "A"),
            (i + 1, j, "A"),
        ]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j, "B"), (*coob, lbl)))

    return sort_unique(edges)


def edges_2d_triangular(Lx, Ly, cyclic=False, cells=None):
    """Return the graph edges of a finite 2D triangular lattice. There is a
    single site per cell, and note the cells do not form a square tiling.
    The nodes (sites) are labelled like ``(i, j)``.

    Parameters
    ----------
    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly))``.

    Returns
    -------
    edges : list[((int, int), (int, int))]
    """
    if cells is None:
        cells = itertools.product(range(Lx), range(Ly))

    edges = []
    for i, j in cells:
        for coob in [(i, j + 1), (i + 1, j), (i + 1, j - 1)]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j), coob))

    return sort_unique(edges)


def edges_2d_triangular_rectangular(Lx, Ly, cyclic=False, cells=None):
    """Return the graph edges of a finite 2D triangular lattice tiled in a
    rectangular geometry. There are two sites per rectangular cell. The nodes
    (sites) are labelled like ``(i, j, s)`` for ``s`` in ``'AB'``.

    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly))``.

    Returns
    -------
    edges : list[((int, int, s), (int, int, s))]
    """
    if cells is None:
        cells = itertools.product(range(Lx), range(Ly))

    edges = []
    for i, j in cells:
        for *coob, lbl in [
            (i, j, "B"),
            (i, j - 1, "B"),
            (i, j + 1, "A"),
        ]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j, "A"), (*coob, lbl)))

        for *coob, lbl in [
            (i + 1, j, "A"),
            (i, j + 1, "B"),
            (i + 1, j + 1, "A"),
        ]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j, "B"), (*coob, lbl)))

    return sort_unique(edges)


def edges_2d_kagome(Lx, Ly, cyclic=False, cells=None):
    """Return the graph edges of a finite 2D kagome lattice. There are
    three sites per cell, and note the cells do not form a square tiling. The
    nodes (sites) are labelled like ``(i, j, s)`` for ``s`` in ``'ABC'``.

    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly))``.

    Returns
    -------
    edges : list[((int, int, str), (int, int, str))]
    """
    if cells is None:
        cells = itertools.product(range(Lx), range(Ly))

    edges = []
    for i, j in cells:
        for *coob, lbl in [
            (i, j, "B"),
            (i, j - 1, "B"),
            (i, j, "C"),
            (i - 1, j, "C"),
        ]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j, "A"), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, "C"),
            (i - 1, j + 1, "C"),
            (i, j, "A"),
            (i, j + 1, "A"),
        ]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j, "B"), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, "A"),
            (i + 1, j, "A"),
            (i, j, "B"),
            (i + 1, j - 1, "B"),
        ]:
            coob = check_2d(coob, Lx, Ly, cyclic)
            if coob:
                edges.append(((i, j, "C"), (*coob, lbl)))

    return sort_unique(edges)


# ----------------------------------- 3D ------------------------------------ #


def check_3d(coo, Lx, Ly, Lz, cyclic):
    """Check ``coo`` in inbounds for a maybe cyclic 3D lattice."""
    x, y, z = coo
    OBC = not cyclic
    inbounds = (0 <= x < Lx) and (0 <= y < Ly) and (0 <= z < Lz)
    if OBC and not inbounds:
        return
    return (x % Lx, y % Ly, z % Lz)


def edges_3d_cubic(Lx, Ly, Lz, cyclic=False, cells=None):
    """Return the graph edges of a finite 3D cubic lattice. The nodes
    (sites) are labelled like ``(i, j, k)``.

    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    Lz : int
        The number of cells along the z-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly), range(Lz))``.

    Returns
    -------
    edges : list[((int, int, int), (int, int, int))]
    """
    if cells is None:
        cells = itertools.product(range(Lx), range(Ly), range(Lz))

    edges = []
    for i, j, k in cells:
        for coob in [(i, j, k + 1), (i, j + 1, k), (i + 1, j, k)]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k), coob))

    return sort_unique(edges)


def edges_3d_pyrochlore(Lx, Ly, Lz, cyclic=False, cells=None):
    """Return the graph edges of a finite 3D pyorchlore lattice. There are
    four sites per cell, and note the cells do not form a cubic tiling. The
    nodes (sites) are labelled like ``(i, j, k, s)`` for ``s`` in ``'ABCD'``.

    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    Lz : int
        The number of cells along the z-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly), range(Lz))``.

    Returns
    -------
    edges : list[((int, int, int, str), (int, int, int, str))]
    """
    if cells is None:
        cells = itertools.product(range(Lx), range(Ly), range(Lz))

    edges = []
    for i, j, k in cells:
        for *coob, lbl in [
            (i, j, k, "B"),
            (i, j - 1, k, "B"),
            (i, j, k, "C"),
            (i - 1, j, k, "C"),
            (i, j, k, "D"),
            (i, j, k - 1, "D"),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, "A"), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, k, "C"),
            (i - 1, j + 1, k, "C"),
            (i, j, k, "D"),
            (i, j + 1, k - 1, "D"),
            (i, j, k, "A"),
            (i, j + 1, k, "A"),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, "B"), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, k, "D"),
            (i + 1, j, k - 1, "D"),
            (i, j, k, "A"),
            (i + 1, j, k, "A"),
            (i, j, k, "B"),
            (i + 1, j - 1, k, "B"),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, "C"), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, k, "A"),
            (i, j, k + 1, "A"),
            (i, j, k, "B"),
            (i, j - 1, k + 1, "B"),
            (i, j, k, "C"),
            (i - 1, j, k + 1, "C"),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, "D"), (*coob, lbl)))

    return sort_unique(edges)


def edges_3d_diamond(Lx, Ly, Lz, cyclic=False, cells=None):
    """Return the graph edges of a finite 3D diamond lattice. There are
    two sites per cell, and note the cells do not form a cubic tiling.  The
    nodes (sites) are labelled like ``(i, j, k, s)`` for ``s`` in ``'AB'``.

    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    Lz : int
        The number of cells along the z-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly), range(Lz))``.

    Returns
    -------
    edges : list[((int, int, int, str), (int, int, int, str))]
    """
    if cells is None:
        cells = itertools.product(range(Lx), range(Ly), range(Lz))

    edges = []
    for i, j, k in cells:
        for *coob, lbl in [
            (i, j, k, "B"),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, "A"), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, k + 1, "A"),
            (i, j + 1, k, "A"),
            (i + 1, j, k, "A"),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, "B"), (*coob, lbl)))

    return sort_unique(edges)


def edges_3d_diamond_cubic(Lx, Ly, Lz, cyclic=False, cells=None):
    """Return the graph edges of a finite 3D diamond lattice tiled in a cubic
    geometry. There are eight sites per cubic cell. The nodes (sites) are
    labelled like ``(i, j, k, s)`` for ``s`` in ``'ABCDEFGH'``.

    Parameters
    ----------
    Lx : int
        The number of cells along the x-direction.
    Ly : int
        The number of cells along the y-direction.
    Lz : int
        The number of cells along the z-direction.
    cyclic : bool, optional
        Whether to use periodic boundary conditions.
    cells : list, optional
        A list of cells to use. If not given the cells used are
        ``itertools.product(range(Lx), range(Ly), range(Lz))``.

    Returns
    -------
    edges : list[((int, int, int, str), (int, int, int, str))]
    """

    if cells is None:
        cells = itertools.product(range(Lx), range(Ly), range(Lz))

    edges = []
    for i, j, k in cells:
        for *coob, lbl in [
            (i, j, k, "E"),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, "A"), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, k, "E"),
            (i, j, k, "F"),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, "B"), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, k, "E"),
            (i, j, k, "G"),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, "C"), (*coob, lbl)))

        for *coob, lbl in [
            (i, j, k, "E"),
            (i, j, k, "H"),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, "D"), (*coob, lbl)))

        for *coob, lbl in []:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, "E"), (*coob, lbl)))

        for *coob, lbl in [
            (i, j + 1, k, "C"),
            (i + 1, j, k, "D"),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, "F"), (*coob, lbl)))

        for *coob, lbl in [
            (i + 1, j, k + 1, "A"),
            (i, j, k + 1, "B"),
            (i + 1, j, k, "D"),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, "G"), (*coob, lbl)))

        for *coob, lbl in [
            (i, j + 1, k + 1, "A"),
            (i, j, k + 1, "B"),
            (i, j + 1, k, "C"),
        ]:
            coob = check_3d(coob, Lx, Ly, Lz, cyclic)
            if coob:
                edges.append(((i, j, k, "H"), (*coob, lbl)))

    return sort_unique(edges)


def edges_tree_rand(n, max_degree=None, seed=None):
    """Return a random tree with ``n`` nodes. This a convenience function for
    testing purposes and the trees generated are not guaranteed to be uniformly
    random (for that see ``networkx.random_labeled_tree``).

    Parameters
    ----------
    n : int
        The number of nodes.
    max_degree : int, optional
        The maximum degree of the nodes. For example ``max_degree=3`` means
        generate a binary tree.
    seed : int, optional
        The random seed.

    Returns
    -------
    edges : list[(int, int)]
    """
    rng = random.Random(seed)
    edges = []

    if max_degree is None:
        nodes = [0]
        for i in range(1, n):
            ib = rng.choice(nodes)
            nodes.append(i)
            edges.append((ib, i))
    else:
        degrees = {0: 0}
        for i in range(1, n):
            ib = rng.choice(list(degrees))
            edges.append((ib, i))

            degrees[i] = 1
            if degrees[ib] + 1 == max_degree:
                # node is finished
                degrees.pop(ib)
            else:
                degrees[ib] += 1

    return edges



import math

import autoray as ar

import quimb.tensor as qtn
from quimb.utils import oset

from .bp_common import (
    BeliefPropagationCommon,
    combine_local_contractions,
    create_lazy_community_edge_map,
)


def _identity(x):
    return x


class L2BP(BeliefPropagationCommon):
    """Lazy (as in multiple uncontracted tensors per site) 2-norm (as in for
    wavefunctions and operators) belief propagation.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to form the 2-norm of and run BP on.
    site_tags : sequence of str, optional
        The tags identifying the sites in ``tn``, each tag forms a region,
        which should not overlap. If the tensor network is structured, then
        these are inferred automatically.
    damping : float or callable, optional
        The damping factor to apply to messages. This simply mixes some part
        of the old message into the new one, with the final message being
        ``damping * old + (1 - damping) * new``. This makes convergence more
        reliable but slower.
    update : {'sequential', 'parallel'}, optional
        Whether to update messages sequentially (newly computed messages are
        immediately used for other updates in the same iteration round) or in
        parallel (all messages are comptued using messages from the previous
        round only). Sequential generally helps convergence but parallel can
        possibly converge to differnt solutions.
    normalize : {'L1', 'L2', 'L2phased', 'Linf', callable}, optional
        How to normalize messages after each update. If None choose
        automatically. If a callable, it should take a message and return the
        normalized message. If a string, it should be one of 'L1', 'L2',
        'L2phased', 'Linf' for the corresponding norms. 'L2phased' is like 'L2'
        but also normalizes the phase of the message, by default used for
        complex dtypes.
    distance : {'L1', 'L2', 'L2phased', 'Linf', 'cosine', callable}, optional
        How to compute the distance between messages to check for convergence.
        If None choose automatically. If a callable, it should take two
        messages and return the distance. If a string, it should be one of
        'L1', 'L2', 'L2phased', 'Linf', or 'cosine' for the corresponding
        norms. 'L2phased' is like 'L2' but also normalizes the phases of the
        messages, by default used for complex dtypes if phased normalization is
        not already being used.
    symmetrize : bool or callable, optional
        Whether to symmetrize the messages, i.e. for each message ensure that
        it is hermitian with respect to its bra and ket indices. If a callable
        it should take a message and return the symmetrized message.
    local_convergence : bool, optional
        Whether to allow messages to locally converge - i.e. if all their
        input messages have converged then stop updating them.
    optimize : str or PathOptimizer, optional
        The path optimizer to use when contracting the messages.
    contract_every : int, optional
        If not None, 'contract' (via BP) the tensor network every
        ``contract_every`` iterations. The resulting values are stored in
        ``zvals`` at corresponding points ``zval_its``.
    inplace : bool, optional
        Whether to perform any operations inplace on the input tensor network.
    contract_opts
        Other options supplied to ``cotengra.array_contract``.
    """

    def __init__(
        self,
        tn,
        site_tags=None,
        *,
        damping=0.0,
        update="sequential",
        normalize=None,
        distance=None,
        symmetrize=True,
        local_convergence=True,
        optimize="auto-hq",
        contract_every=None,
        inplace=False,
        **contract_opts,
    ):
        super().__init__(
            tn,
            damping=damping,
            update=update,
            normalize=normalize,
            distance=distance,
            contract_every=contract_every,
            inplace=inplace,
        )

        self.local_convergence = local_convergence
        self.optimize = optimize
        self.contract_opts = contract_opts

        if site_tags is None:
            self.site_tags = tuple(tn.site_tags)
        else:
            self.site_tags = tuple(site_tags)

        (
            self.edges,
            self.neighbors,
            self.local_tns,
            self.touch_map,
        ) = create_lazy_community_edge_map(tn, site_tags)
        self.touched = oset()

        # these are all settable properties
        self.symmetrize = symmetrize

        # initialize messages
        self.messages = {}

        for pair, bix in self.edges.items():
            cix = tuple(ix + "_l2bp*" for ix in bix)
            remapper = dict(zip(bix, cix))
            output_inds = cix + bix

            # compute leftwards and righwards messages
            for i, j in (sorted(pair), sorted(pair, reverse=True)):
                tn_i = self.local_tns[i]
                tn_i2 = tn_i & tn_i.conj().reindex_(remapper)
                tm = tn_i2.contract(
                    all,
                    output_inds=output_inds,
                    optimize=self.optimize,
                    drop_tags=True,
                    **self.contract_opts,
                )
                tm.modify(apply=self._symmetrize_fn)
                tm.modify(apply=self._normalize_fn)
                self.messages[i, j] = tm

        # initialize contractions
        self.contraction_tns = {}
        for pair, bix in self.edges.items():
            for i, j in (sorted(pair), sorted(pair, reverse=True)):
                # form the ket side and messages
                tn_i_left = self.local_tns[i]
                # get other incident nodes which aren't j
                ks = [k for k in self.neighbors[i] if k != j]
                tks = [self.messages[k, i] for k in ks]

                # form the 'bra' side
                tn_i_right = tn_i_left.conj()
                # get the bonds that attach the bra to messages
                outer_bix = {
                    ix for k in ks for ix in self.edges[tuple(sorted((k, i)))]
                }
                # need to reindex to join message bonds, and create bra outputs
                remapper = {}
                for ix in tn_i_right.ind_map:
                    if ix in bix:
                        # bra outputs
                        remapper[ix] = ix + "_l2bp**"
                    elif ix in outer_bix:
                        # messages connected
                        remapper[ix] = ix + "_l2bp*"
                    # remaining indices are either internal and will be mangled
                    # or global outer indices and will be contracted directly

                tn_i_right.reindex_(remapper)

                self.contraction_tns[i, j] = qtn.TensorNetwork(
                    (tn_i_left, *tks, tn_i_right), virtual=True
                )

    @property
    def symmetrize(self):
        return self._symmetrize

    @symmetrize.setter
    def symmetrize(self, symmetrize):
        if callable(symmetrize):
            # explicit function
            self._symmetrize = True
            self._symmetrize_fn = symmetrize

        elif symmetrize:
            # default symmetrization
            _transpose = ar.get_lib_fn(self.backend, "transpose")
            _conj = ar.get_lib_fn(self.backend, "conj")

            def _symmetrize_fn(x):
                N = ar.ndim(x)
                perm = (*range(N // 2, N), *range(0, N // 2))
                # XXX: do this blockwise for block/fermi arrays?
                return x + _conj(_transpose(x, perm))

            self._symmetrize = True
            self._symmetrize_fn = _symmetrize_fn

        else:
            # no symmetrization
            self._symmetrize = False
            self._symmetrize_fn = _identity

    def iterate(self, tol=5e-6):
        if (not self.local_convergence) or (not self.touched):
            # assume if asked to iterate that we want to check all messages
            self.touched.update(
                pair for edge in self.edges for pair in (edge, edge[::-1])
            )

        ncheck = len(self.touched)
        nconv = 0
        max_mdiff = -1.0
        new_touched = oset()

        def _compute_m(key):
            i, j = key
            bix = self.edges[(i, j) if i < j else (j, i)]
            cix = tuple(ix + "_l2bp**" for ix in bix)
            output_inds = cix + bix

            tn_i_to_j = self.contraction_tns[i, j]

            tm_new = tn_i_to_j.contract(
                all,
                output_inds=output_inds,
                drop_tags=True,
                optimize=self.optimize,
                **self.contract_opts,
            )
            tm_new.modify(apply=self._symmetrize_fn)
            tm_new.modify(apply=self._normalize_fn)
            return tm_new.data

        def _update_m(key, data):
            nonlocal nconv, max_mdiff

            tm = self.messages[key]

            # pre-damp distance
            mdiff = self._distance_fn(data, tm.data)

            if self.damping:
                data = self._damping_fn(data, tm.data)

            # # post-damp distance
            # mdiff = self._distance_fn(data, tm.data)

            if mdiff > tol:
                # mark touching messages for update
                new_touched.update(self.touch_map[key])
            else:
                nconv += 1

            max_mdiff = max(max_mdiff, mdiff)
            tm.modify(data=data)

        if self.update == "parallel":
            new_data = {}
            # compute all new messages
            while self.touched:
                key = self.touched.pop()
                new_data[key] = _compute_m(key)
            # insert all new messages
            for key, data in new_data.items():
                _update_m(key, data)

        elif self.update == "sequential":
            # compute each new message and immediately re-insert it
            while self.touched:
                key = self.touched.pop()
                data = _compute_m(key)
                _update_m(key, data)

        self.touched = new_touched

        return {
            "nconv": nconv,
            "ncheck": ncheck,
            "max_mdiff": max_mdiff,
        }

    def normalize_message_pairs(self):
        """Normalize all messages such that for each bond `<m_i|m_j> = 1` and
        `<m_i|m_i> = <m_j|m_j>` (but in general != 1). This is different to
        normalizing each message.
        """
        for i, j in self.edges:
            tmi = self.messages[i, j]
            tmj = self.messages[j, i]
            nij = (tmi @ tmj) ** 0.5
            nii = (tmi @ tmi) ** 0.25
            njj = (tmj @ tmj) ** 0.25
            tmi /= nij * nii / njj
            tmj /= nij * njj / nii

    def contract(self, strip_exponent=False, check_zero=True):
        """Estimate the contraction of the norm squared using the current
        messages.
        """
        zvals = []
        for i, ket in self.local_tns.items():
            # we allow missing keys here for tensors which are just
            # disconnected but still appear in local_tns
            ks = self.neighbors.get(i, ())
            bix = [ix for k in ks for ix in self.edges[tuple(sorted((k, i)))]]
            bra = ket.H.reindex_({ix: ix + "_l2bp*" for ix in bix})
            tni = qtn.TensorNetwork(
                (
                    ket,
                    *(self.messages[k, i] for k in ks),
                    bra,
                )
            )
            z = tni.contract(all, optimize=self.optimize, **self.contract_opts)
            zvals.append((z, 1))

        for i, j in self.edges:
            z = (self.messages[i, j] & self.messages[j, i]).contract(
                all,
                optimize=self.optimize,
                **self.contract_opts,
            )
            # power / counting factor is -1 for messages, i.e. divide
            zvals.append((z, -1))

        return combine_local_contractions(
            zvals,
            backend=self.backend,
            strip_exponent=strip_exponent,
            check_zero=check_zero,
        )

    def partial_trace(
        self,
        site,
        normalized=True,
        optimize="auto-hq",
    ):
        example_tn = next(tn for tn in self.local_tns.values())

        site_tag = example_tn.site_tag(site)
        ket_site_ind = example_tn.site_ind(site)

        ks = self.neighbors[site_tag]
        tn_rho_i = self.local_tns[site_tag].copy()
        tn_bra_i = tn_rho_i.H

        for k in ks:
            tn_rho_i &= self.messages[k, site_tag]

        outer_bix = {
            ix for k in ks for ix in self.edges[tuple(sorted((k, site_tag)))]
        }

        ind_changes = {}
        for ix in tn_bra_i.ind_map:
            if ix == ket_site_ind:
                # open up the site index
                bra_site_ind = ix + "_l2bp**"
                ind_changes[ix] = bra_site_ind
            if ix in outer_bix:
                # attach bra message indices
                ind_changes[ix] = ix + "_l2bp*"
        tn_bra_i.reindex_(ind_changes)

        tn_rho_i &= tn_bra_i

        rho_i = tn_rho_i.to_dense(
            [ket_site_ind],
            [bra_site_ind],
            optimize=optimize,
            **self.contract_opts,
        )
        if normalized:
            rho_i = rho_i / ar.do("trace", rho_i)

        return rho_i

    def compress(
        self,
        tn,
        max_bond=None,
        cutoff=5e-6,
        cutoff_mode="rsum2",
        renorm=0,
        lazy=False,
    ):
        """Compress the state ``tn``, assumed to matched this L2BP instance,
        using the messages stored.
        """
        for (i, j), bix in self.edges.items():
            tml = self.messages[i, j]
            tmr = self.messages[j, i]

            bix_sizes = [tml.ind_size(ix) for ix in bix]
            dm = math.prod(bix_sizes)

            ml = ar.reshape(tml.data, (dm, dm))
            dl = self.local_tns[i].outer_size() // dm
            Rl = qtn.decomp.squared_op_to_reduced_factor(
                ml, dl, dm, right=True
            )

            mr = ar.reshape(tmr.data, (dm, dm)).T
            dr = self.local_tns[j].outer_size() // dm
            Rr = qtn.decomp.squared_op_to_reduced_factor(
                mr, dm, dr, right=False
            )

            Pl, Pr = qtn.decomp.compute_oblique_projectors(
                Rl,
                Rr,
                cutoff_mode=cutoff_mode,
                renorm=renorm,
                max_bond=max_bond,
                cutoff=cutoff,
            )

            Pl = ar.do("reshape", Pl, (*bix_sizes, -1))
            Pr = ar.do("reshape", Pr, (-1, *bix_sizes))

            ltn = tn.select(i)
            rtn = tn.select(j)

            new_lix = [qtn.rand_uuid() for _ in bix]
            new_rix = [qtn.rand_uuid() for _ in bix]
            new_bix = [qtn.rand_uuid()]
            ltn.reindex_(dict(zip(bix, new_lix)))
            rtn.reindex_(dict(zip(bix, new_rix)))

            # ... and insert the new projectors in place
            tn |= qtn.Tensor(Pl, inds=new_lix + new_bix, tags=(i,))
            tn |= qtn.Tensor(Pr, inds=new_bix + new_rix, tags=(j,))

        if not lazy:
            for st in self.site_tags:
                try:
                    tn.contract_tags_(
                        st, optimize=self.optimize, **self.contract_opts
                    )
                except KeyError:
                    pass

        return tn


def contract_l2bp(
    tn,
    site_tags=None,
    damping=0.0,
    update="sequential",
    local_convergence=True,
    optimize="auto-hq",
    max_iterations=1000,
    tol=5e-6,
    strip_exponent=False,
    info=None,
    progbar=False,
    **contract_opts,
):
    """Estimate the norm squared of ``tn`` using lazy belief propagation.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to estimate the norm squared of.
    site_tags : sequence of str, optional
        The tags identifying the sites in ``tn``, each tag forms a region.
    damping : float, optional
        The damping parameter to use, defaults to no damping.
    update : {'parallel', 'sequential'}, optional
        Whether to update all messages in parallel or sequentially.
    local_convergence : bool, optional
        Whether to allow messages to locally converge - i.e. if all their
        input messages have converged then stop updating them.
    optimize : str or PathOptimizer, optional
        The contraction strategy to use.
    max_iterations : int, optional
        The maximum number of iterations to perform.
    tol : float, optional
        The convergence tolerance for messages.
    strip_exponent : bool, optional
        Whether to strip the exponent from the final result. If ``True``
        then the returned result is ``(mantissa, exponent)``.
    info : dict, optional
        If specified, update this dictionary with information about the
        belief propagation run.
    progbar : bool, optional
        Whether to show a progress bar.
    contract_opts
        Other options supplied to ``cotengra.array_contract``.
    """
    bp = L2BP(
        tn,
        site_tags=site_tags,
        damping=damping,
        update=update,
        local_convergence=local_convergence,
        optimize=optimize,
        **contract_opts,
    )
    bp.run(
        max_iterations=max_iterations,
        tol=tol,
        info=info,
        progbar=progbar,
    )
    return bp.contract(strip_exponent=strip_exponent)


def compress_l2bp(
    tn,
    max_bond,
    cutoff=0.0,
    cutoff_mode="rsum2",
    max_iterations=1000,
    tol=5e-6,
    site_tags=None,
    damping=0.0,
    update="sequential",
    local_convergence=True,
    optimize="auto-hq",
    lazy=False,
    inplace=False,
    info=None,
    progbar=False,
    **contract_opts,
):
    """Compress ``tn`` using lazy belief propagation, producing a tensor
    network with a single tensor per site.

    Parameters
    ----------
    tn : TensorNetwork
        The tensor network to form the 2-norm of, run BP on and then compress.
    max_bond : int
        The maximum bond dimension to compress to.
    cutoff : float, optional
        The cutoff to use when compressing.
    cutoff_mode : int, optional
        The cutoff mode to use when compressing.
    max_iterations : int, optional
        The maximum number of iterations to perform.
    tol : float, optional
        The convergence tolerance for messages.
    site_tags : sequence of str, optional
        The tags identifying the sites in ``tn``, each tag forms a region. If
        the tensor network is structured, then these are inferred
        automatically.
    damping : float, optional
        The damping parameter to use, defaults to no damping.
    update : {'parallel', 'sequential'}, optional
        Whether to update all messages in parallel or sequentially.
    local_convergence : bool, optional
        Whether to allow messages to locally converge - i.e. if all their
        input messages have converged then stop updating them.
    optimize : str or PathOptimizer, optional
        The path optimizer to use when contracting the messages.
    lazy : bool, optional
        Whether to perform the compression lazily, i.e. to leave the computed
        compression projectors uncontracted.
    inplace : bool, optional
        Whether to perform the compression inplace.
    info : dict, optional
        If specified, update this dictionary with information about the
        belief propagation run.
    progbar : bool, optional
        Whether to show a progress bar.
    contract_opts
        Other options supplied to ``cotengra.array_contract``.

    Returns
    -------
    TensorNetwork
    """
    tnc = tn if inplace else tn.copy()

    bp = L2BP(
        tnc,
        site_tags=site_tags,
        damping=damping,
        update=update,
        local_convergence=local_convergence,
        optimize=optimize,
        **contract_opts,
    )
    bp.run(
        max_iterations=max_iterations,
        tol=tol,
        info=info,
        progbar=progbar,
    )
    return bp.compress(
        tnc,
        max_bond=max_bond,
        cutoff=cutoff,
        cutoff_mode=cutoff_mode,
        lazy=lazy,
    )



"""Decorator for automatically just in time compiling tensor network functions.

TODO:
- [ ] check and cache on input shapes
"""

import functools

import autoray as ar

from quimb.tensor.interface import pack, unpack


def try_and_get_params(x):
    if hasattr(x, "get_params"):
        return x.get_params()
    return x


class AutojittedTN:
    def __init__(self, fn, decorator=ar.autojit, **decorator_opts):
        self.fn = fn
        self.jit_fn = None
        self.decorator = decorator
        self.decorator_opts = decorator_opts

    def _setup(self, *args, **kwargs):
        # extract all arrays to flat list, keep reference pytree and skeletons
        flat, ref_tree = ar.tree_flatten((args, kwargs), get_ref=True)
        _, skeletons = zip(*map(pack, flat))

        # now decorate the function that takes pytrees of arrays only

        @self.decorator(**self.decorator_opts)
        def pyfn(*pytrees):
            # inject back into the skeletons
            flat = [unpack(p, skel) for p, skel in zip(pytrees, skeletons)]
            args, kwargs = ar.tree_unflatten(flat, ref_tree)

            # call the original function
            result = self.fn(*args, **kwargs)

            # flatten the result to check if it needs unpacking
            flat_out, ref_tree_out = ar.tree_flatten(result, get_ref=True)

            if not any(hasattr(x, "set_params") for x in flat_out):
                # result doesnt need unpacking -> is a pure pytree
                self.ref_tree_out = None
                self.skeletons_out = None
                return result

            # need to unpack after each call
            params_out, skeletons_out = zip(*map(pack, flat_out))
            self.skeletons_out = skeletons_out
            self.ref_tree_out = ref_tree_out
            return params_out

        self.jit_fn = pyfn

    def __call__(self, *args, backend=None, **kwargs):
        if self.jit_fn is None:
            self._setup(*args, **kwargs)

        # extract to sequence of pytrees of arrays
        pytrees = tuple(
            try_and_get_params(x) for x in ar.tree_flatten((args, kwargs))
        )

        # call the traced function
        result = self.jit_fn(*pytrees, backend=backend)

        if self.skeletons_out is None:
            # no need to unpack
            return result

        # unpack into tensor networks instances etc
        flat = [unpack(p, skel) for p, skel in zip(result, self.skeletons_out)]
        return ar.tree_unflatten(flat, self.ref_tree_out)


def autojit_tn(
    fn=None,
    decorator=ar.autojit,
    **decorator_opts,
):
    """Decorate a tensor network function to be just in time compiled / traced.
    This traces solely array operations resulting in a completely static
    computational graph with no side-effects. The resulting function can be
    much faster if called repeatedly with only numeric changes, or hardware
    accelerated if a library such as ``jax`` is used.

    Parameters
    ----------
    fn : callable
        The function to be decorated.
    decorator : callable
        The decorator to use to wrap the underlying array function. For example
        ``jax.jit``. Defaults to ``autoray.autojit``.
    decorator_opts
        Options to pass to the decorator, e.g. ``backend`` for
        ``autoray.autojit``.
    """
    kwargs = {
        "decorator": decorator,
        **decorator_opts,
    }
    if fn is None:
        return functools.partial(autojit_tn, **kwargs)
    return functools.wraps(fn)(AutojittedTN(fn, **kwargs))



import pytest

import quimb as qu
import quimb.tensor as qtn
import quimb.tensor.belief_propagation as qbp


@pytest.mark.parametrize("dtype", ["float32", "complex64"])
@pytest.mark.parametrize("local_convergence", [False, True])
@pytest.mark.parametrize("normalize", ["L1", "L2", "Linf"])
def test_contract_tree_exact(dtype, local_convergence, normalize):
    tn = qtn.TN_rand_tree(10, 3, seed=42, dtype=dtype)
    Z_ex = tn.contract()
    info = {}
    Z_bp = qbp.contract_l1bp(
        tn,
        info=info,
        normalize=normalize,
        local_convergence=local_convergence,
        progbar=True,
    )
    assert info["converged"]
    assert Z_ex == pytest.approx(Z_bp, rel=5e-6)


@pytest.mark.parametrize("dtype", ["float32", "complex64"])
@pytest.mark.parametrize("damping", [0.0, 0.1])
@pytest.mark.parametrize("diis", [False, True])
def test_contract_loopy_approx(dtype, damping, diis):
    tn = qtn.TN2D_rand(3, 4, 5, dtype=dtype, dist="uniform")
    Z_ex = tn.contract()
    info = {}
    Z_bp = qbp.contract_l1bp(
        tn,
        damping=damping,
        diis=diis,
        info=info,
        progbar=True,
    )
    assert info["converged"]
    assert Z_ex == pytest.approx(Z_bp, rel=0.1)


@pytest.mark.parametrize("dtype", ["float32", "complex64"])
@pytest.mark.parametrize("damping", [0.0, 0.1])
@pytest.mark.parametrize("update", ("parallel", "sequential"))
def test_contract_double_loopy_approx(dtype, damping, update):
    peps = qtn.PEPS.rand(4, 3, 2, seed=42, dtype=dtype)
    tn = peps.H & peps
    Z_ex = tn.contract()
    info = {}
    Z_bp1 = qbp.contract_l1bp(
        tn,
        damping=damping,
        update=update,
        info=info,
        progbar=True,
    )
    assert info["converged"]
    assert Z_bp1 == pytest.approx(Z_ex, rel=0.3)
    # compare with 2-norm BP on the peps directly
    Z_bp2 = qbp.contract_d2bp(peps)
    assert Z_bp1 == pytest.approx(Z_bp2, rel=5e-5)


@pytest.mark.parametrize("dtype", ["float32", "complex64"])
def test_contract_tree_triple_sandwich_exact(dtype):
    edges = qtn.edges_tree_rand(20, 3, seed=42)
    ket = qtn.TN_from_edges_rand(
        edges,
        3,
        phys_dim=2,
        seed=42,
        site_ind_id="k{}",
        dtype=dtype,
    )
    op = qtn.TN_from_edges_rand(
        edges,
        2,
        phys_dim=2,
        seed=42,
        site_ind_id=("k{}", "b{}"),
        dtype=dtype,
    )
    bra = qtn.TN_from_edges_rand(
        edges,
        3,
        phys_dim=2,
        seed=42,
        site_ind_id="b{}",
        dtype=dtype,
    )
    tn = bra.H | op | ket
    Z_ex = tn.contract()
    info = {}
    Z_bp = qbp.contract_l1bp(tn, info=info, progbar=True)
    assert info["converged"]
    assert Z_ex == pytest.approx(Z_bp, rel=5e-6)


@pytest.mark.parametrize("dtype", ["float32", "complex64"])
@pytest.mark.parametrize("damping", [0.0, 0.1])
@pytest.mark.parametrize("diis", [False, True])
def test_contract_tree_triple_sandwich_loopy_approx(dtype, damping, diis):
    edges = qtn.edges_2d_hexagonal(2, 3)
    ket = qtn.TN_from_edges_rand(
        edges,
        3,
        phys_dim=2,
        seed=42,
        site_ind_id="k{}",
        dtype=dtype,
        # make the wavefunction postive to make easier
        dist="uniform",
    )
    ket /= (ket.H @ ket) ** 0.5

    G_ket = ket.gate(qu.pauli("Z"), [(1, 1, "A")], propagate_tags="sites")
    tn = ket.H | G_ket
    Z_ex = tn.contract()
    info = {}
    Z_bp = qbp.contract_l1bp(
        tn,
        damping=damping,
        diis=diis,
        info=info,
        progbar=True,
    )
    assert info["converged"]
    assert Z_bp == pytest.approx(Z_ex, rel=0.5)


def test_contract_cluster_approx():
    tn = qtn.TN2D_classical_ising_partition_function(8, 8, 0.4, h=0.2)
    f_ex = qu.log(tn.contract())
    f_bp = qu.log(qbp.contract_l1bp(tn))
    assert f_bp == pytest.approx(f_ex, rel=0.3)
    cluster_tags = []
    for i in range(0, 8, 2):
        for j in range(0, 8, 2):
            cluster_tag = f"C{i},{j}"
            tn[i, j].add_tag(cluster_tag)
            tn[i, j + 1].add_tag(cluster_tag)
            tn[i + 1, j].add_tag(cluster_tag)
            tn[i + 1, j + 1].add_tag(cluster_tag)
            cluster_tags.append(cluster_tag)
    info = {}
    f_bp2 = qu.log(
        qbp.contract_l1bp(tn, site_tags=cluster_tags, info=info, progbar=True)
    )
    assert info["converged"]
    assert f_bp == pytest.approx(f_ex, rel=0.1)
    assert abs(1 - f_ex / f_bp2) < abs(1 - f_ex / f_bp)


def test_mps():
    # catch bug to do with structured contract and output inds
    L = 6
    psi = qtn.MPS_rand_state(L=L, seed=20, bond_dim=3)
    psiG = psi.copy()
    psiG.gate_(qu.pauli("X"), 5, contract=True)
    expec = psi.H & psiG
    O = qbp.contract_l1bp(
        expec,
        site_tags=[f"I{i}" for i in range(L)],
    )
    assert O == pytest.approx(expec ^ ..., abs=1e-6)



import pytest

import quimb as qu
import quimb.tensor as qtn
import quimb.tensor.belief_propagation as qbp


@pytest.mark.parametrize("damping", [0.0, 0.1, 0.5])
@pytest.mark.parametrize("diis", [False, True])
def test_contract_hyper(damping, diis):
    htn = qtn.HTN_random_ksat(3, 50, alpha=2.0, seed=42, mode="dense")
    info = {}
    num_solutions = qbp.contract_hv1bp(
        htn, damping=damping, diis=diis, info=info, progbar=True
    )
    assert info["converged"]
    assert num_solutions == pytest.approx(309273226, rel=0.1)


@pytest.mark.parametrize("messages", [None, "dense", "random"])
def test_contract_tree_exact(messages):
    tn = qtn.TN_rand_tree(20, 3)
    Z = tn.contract()
    info = {}

    if messages == "random":

        def messages(shape):
            return qu.randn(shape, dist="uniform")

    Z_bp = qbp.contract_hv1bp(tn, messages=messages, info=info, progbar=True)
    assert info["converged"]
    assert Z == pytest.approx(Z_bp, rel=1e-12)


@pytest.mark.parametrize("damping", [0.0, 0.1, 0.5])
@pytest.mark.parametrize("diis", [False, True])
def test_contract_normal(damping, diis):
    tn = qtn.TN2D_from_fill_fn(lambda s: qu.randn(s, dist="uniform"), 6, 6, 2)
    Z = tn.contract()
    info = {}
    Z_bp = qbp.contract_hv1bp(
        tn, damping=damping, diis=diis, info=info, progbar=True
    )
    assert info["converged"]
    assert Z == pytest.approx(Z_bp, rel=1e-1)


@pytest.mark.parametrize("damping", [0.0, 0.1])
def test_sample(damping):
    nvars = 20
    htn = qtn.HTN_random_ksat(3, nvars, alpha=2.0, seed=42, mode="dense")
    config, tn_config, omega = qbp.sample_hv1bp(
        htn, damping=damping, seed=42, progbar=True
    )
    assert len(config) == nvars
    assert tn_config.num_indices == 0
    assert tn_config.contract() == pytest.approx(1.0)
    assert 0.0 < omega < 1.0



import importlib

import numpy as np
import pytest

requires_symmray = pytest.mark.skipif(
    importlib.util.find_spec("symmray") is None,
    reason="symmray not installed",
)


@requires_symmray
@pytest.fixture
def get_fpeps_and_norm():
    import symmray as sr

    # fPEPS parameters
    Lx = int(4)
    Ly = int(4)
    symmetry = "Z2"
    D = 4
    # Load PEPS
    fpeps = sr.PEPS_fermionic_rand(
        symmetry, Lx, Ly, bond_dim=D, phys_dim=4, seed=42
    )
    fpeps.equalize_norms_()
    for ts in fpeps.tensors:
        ts.data.phase_sync(inplace=True)
    fpeps_norm = fpeps.make_norm()
    # benchmark_norm = fpeps_norm.contract_boundary_from_xmax(xrange=(0, Lx-1), max_bond=256, cutoff=0.0, mode='direct').contract()
    benchmark_norm = np.float64(9.347604511732736e18)
    return fpeps_norm, benchmark_norm


@requires_symmray
@pytest.mark.parametrize("from_which", ["xmin", "xmax", "ymin", "ymax"])
def test_fmps_mpo_fitting(from_which, get_fpeps_and_norm):
    fpeps_norm, benchmark_norm = get_fpeps_and_norm
    print(f"Benchmark norm: {benchmark_norm}")
    print("Boundary fMPS-MPO fitting contraction test:")
    # contraction bond dimension
    chi = 128
    if from_which == "xmin":
        print("xmin:")
        c_xmin_0 = fpeps_norm.contract_boundary_from_xmin(
            xrange=(0, 1),
            max_bond=chi,
            cutoff=0.0,
            mode="fit",
            tol=1e-5,
            tn_fit="zipup",
            bsz=2,
            max_iterations=6,
        ).contract()
        c_xmin_1 = fpeps_norm.contract_boundary_from_xmin(
            xrange=(0, 2),
            max_bond=chi,
            cutoff=0.0,
            mode="fit",
            tol=1e-5,
            tn_fit="zipup",
            bsz=2,
            max_iterations=6,
        ).contract()
        c_xmin_2 = fpeps_norm.contract_boundary_from_xmin(
            xrange=(0, 1),
            max_bond=chi,
            cutoff=0.0,
            mode="fit",
            tol=1e-5,
            tn_fit="zipup",
            bsz=2,
            max_iterations=5,
        ).contract()
        c_xmin_3 = fpeps_norm.contract_boundary_from_xmin(
            xrange=(0, 2),
            max_bond=chi,
            cutoff=0.0,
            mode="fit",
            tol=1e-5,
            tn_fit="zipup",
            bsz=2,
            max_iterations=5,
        ).contract()
        print(c_xmin_0, np.allclose(c_xmin_0, benchmark_norm, rtol=1e-4))
        print(c_xmin_1, np.allclose(c_xmin_1, benchmark_norm, rtol=1e-4))
        print(c_xmin_2, np.allclose(c_xmin_2, benchmark_norm, rtol=1e-4))
        print(c_xmin_3, np.allclose(c_xmin_3, benchmark_norm, rtol=1e-4))
        assert np.allclose(c_xmin_0, benchmark_norm, rtol=1e-4)
        assert np.allclose(c_xmin_1, benchmark_norm, rtol=1e-4)
        assert np.allclose(c_xmin_2, benchmark_norm, rtol=1e-4)
        assert np.allclose(c_xmin_3, benchmark_norm, rtol=1e-4)
    elif from_which == "xmax":
        print("xmax:")
        c_xmax_0 = fpeps_norm.contract_boundary_from_xmax(
            xrange=(2, 3),
            max_bond=chi,
            cutoff=0.0,
            mode="fit",
            tol=1e-5,
            tn_fit="zipup",
            bsz=2,
            max_iterations=6,
        ).contract()
        c_xmax_1 = fpeps_norm.contract_boundary_from_xmax(
            xrange=(1, 3),
            max_bond=chi,
            cutoff=0.0,
            mode="fit",
            tol=1e-5,
            tn_fit="zipup",
            bsz=2,
            max_iterations=6,
        ).contract()
        c_xmax_2 = fpeps_norm.contract_boundary_from_xmax(
            xrange=(2, 3),
            max_bond=chi,
            cutoff=0.0,
            mode="fit",
            tol=1e-5,
            tn_fit="zipup",
            bsz=2,
            max_iterations=5,
        ).contract()
        c_xmax_3 = fpeps_norm.contract_boundary_from_xmax(
            xrange=(1, 3),
            max_bond=chi,
            cutoff=0.0,
            mode="fit",
            tol=1e-5,
            tn_fit="zipup",
            bsz=2,
            max_iterations=5,
        ).contract()
        print(c_xmax_0, np.allclose(c_xmax_0, benchmark_norm, rtol=1e-4))
        print(c_xmax_1, np.allclose(c_xmax_1, benchmark_norm, rtol=1e-4))
        print(c_xmax_2, np.allclose(c_xmax_2, benchmark_norm, rtol=1e-4))
        print(c_xmax_3, np.allclose(c_xmax_3, benchmark_norm, rtol=1e-4))
        assert np.allclose(c_xmax_0, benchmark_norm, rtol=1e-4)
        assert np.allclose(c_xmax_1, benchmark_norm, rtol=1e-4)
        assert np.allclose(c_xmax_2, benchmark_norm, rtol=1e-4)
        assert np.allclose(c_xmax_3, benchmark_norm, rtol=1e-4)
    elif from_which == "ymin":
        print("ymin:")
        c_ymin_0 = fpeps_norm.contract_boundary_from_ymin(
            yrange=(0, 1),
            max_bond=chi,
            cutoff=0.0,
            mode="fit",
            tol=1e-5,
            tn_fit="zipup",
            bsz=2,
            max_iterations=6,
        ).contract()
        c_ymin_1 = fpeps_norm.contract_boundary_from_ymin(
            yrange=(0, 2),
            max_bond=chi,
            cutoff=0.0,
            mode="fit",
            tol=1e-5,
            tn_fit="zipup",
            bsz=2,
            max_iterations=6,
        ).contract()
        c_ymin_2 = fpeps_norm.contract_boundary_from_ymin(
            yrange=(0, 1),
            max_bond=chi,
            cutoff=0.0,
            mode="fit",
            tol=1e-5,
            tn_fit="zipup",
            bsz=2,
            max_iterations=5,
        ).contract()
        c_ymin_3 = fpeps_norm.contract_boundary_from_ymin(
            yrange=(0, 2),
            max_bond=chi,
            cutoff=0.0,
            mode="fit",
            tol=1e-5,
            tn_fit="zipup",
            bsz=2,
            max_iterations=5,
        ).contract()
        print(c_ymin_0, np.allclose(c_ymin_0, benchmark_norm, rtol=1e-4))
        print(c_ymin_1, np.allclose(c_ymin_1, benchmark_norm, rtol=1e-4))
        print(c_ymin_2, np.allclose(c_ymin_2, benchmark_norm, rtol=1e-4))
        print(c_ymin_3, np.allclose(c_ymin_3, benchmark_norm, rtol=1e-4))
        assert np.allclose(c_ymin_0, benchmark_norm, rtol=1e-4)
        assert np.allclose(c_ymin_1, benchmark_norm, rtol=1e-4)
        assert np.allclose(c_ymin_2, benchmark_norm, rtol=1e-4)
        assert np.allclose(c_ymin_3, benchmark_norm, rtol=1e-4)
    elif from_which == "ymax":
        print("ymax:")
        c_ymax_0 = fpeps_norm.contract_boundary_from_ymax(
            yrange=(2, 3),
            max_bond=chi,
            cutoff=0.0,
            mode="fit",
            tol=1e-5,
            tn_fit="zipup",
            bsz=2,
            max_iterations=6,
        ).contract()
        c_ymax_1 = fpeps_norm.contract_boundary_from_ymax(
            yrange=(1, 3),
            max_bond=chi,
            cutoff=0.0,
            mode="fit",
            tol=1e-5,
            tn_fit="zipup",
            bsz=2,
            max_iterations=6,
        ).contract()
        c_ymax_2 = fpeps_norm.contract_boundary_from_ymax(
            yrange=(2, 3),
            max_bond=chi,
            cutoff=0.0,
            mode="fit",
            tol=1e-5,
            tn_fit="zipup",
            bsz=2,
            max_iterations=5,
        ).contract()
        c_ymax_3 = fpeps_norm.contract_boundary_from_ymax(
            yrange=(1, 3),
            max_bond=chi,
            cutoff=0.0,
            mode="fit",
            tol=1e-5,
            tn_fit="zipup",
            bsz=2,
            max_iterations=5,
        ).contract()
        print(c_ymax_0, np.allclose(c_ymax_0, benchmark_norm, rtol=1e-4))
        print(c_ymax_1, np.allclose(c_ymax_1, benchmark_norm, rtol=1e-4))
        print(c_ymax_2, np.allclose(c_ymax_2, benchmark_norm, rtol=1e-4))
        print(c_ymax_3, np.allclose(c_ymax_3, benchmark_norm, rtol=1e-4))
        assert np.allclose(c_ymax_0, benchmark_norm, rtol=1e-4)
        assert np.allclose(c_ymax_1, benchmark_norm, rtol=1e-4)
        assert np.allclose(c_ymax_2, benchmark_norm, rtol=1e-4)
        assert np.allclose(c_ymax_3, benchmark_norm, rtol=1e-4)



"""Backend agnostic functions for solving matrices either fully or partially."""

import functools
import warnings

import numpy as np
import scipy.linalg as sla
import scipy.sparse.linalg as spla

from ..utils import raise_cant_find_library_function
from ..core import qarray, dag, issparse, isdense, vdot, ldmul
from .numpy_linalg import (
    eig_numpy,
    eigs_numpy,
    svds_numpy,
)

from .scipy_linalg import (
    eigs_scipy,
    eigs_lobpcg,
    eigs_primme,
    svds_scipy,
    svds_primme,
)
from . import SLEPC4PY_FOUND

if SLEPC4PY_FOUND:
    from .mpi_launcher import (
        eigs_slepc_spawn,
        mfn_multiply_slepc_spawn,
        svds_slepc_spawn,
    )
    from .slepc_linalg import eigs_slepc, svds_slepc, mfn_multiply_slepc
else:  # pragma: no cover
    eigs_slepc = raise_cant_find_library_function("slepc4py")
    eigs_slepc_spawn = raise_cant_find_library_function("slepc4py")
    svds_slepc = raise_cant_find_library_function("slepc4py")
    svds_slepc_spawn = raise_cant_find_library_function("slepc4py")
    mfn_multiply_slepc = raise_cant_find_library_function("slepc4py")
    mfn_multiply_slepc_spawn = raise_cant_find_library_function("slepc4py")


# --------------------------------------------------------------------------- #
#                          Partial eigendecomposition                         #
# --------------------------------------------------------------------------- #


def choose_backend(A, k, int_eps=False, B=None):
    """Pick a backend automatically for partial decompositions."""
    # LinOps -> not possible to simply convert to dense or use MPI processes
    A_is_linop = isinstance(A, spla.LinearOperator)
    B_is_linop = isinstance(B, spla.LinearOperator)

    # small array or large part of subspace requested
    small_d_big_k = A.shape[0] ** 2 / k < (10000 if int_eps else 2000)

    if small_d_big_k and not (A_is_linop or B_is_linop):
        return "NUMPY"

    # slepc seems faster for sparse, dense and LinearOperators
    if SLEPC4PY_FOUND and not B_is_linop:
        # only spool up an mpi pool for big sparse matrices though
        if issparse(A) and A.nnz > 10000:
            return "SLEPC"

        return "SLEPC-NOMPI"

    return "SCIPY"


_EIGS_METHODS = {
    "NUMPY": eigs_numpy,
    "SCIPY": eigs_scipy,
    "PRIMME": eigs_primme,
    "LOBPCG": eigs_lobpcg,
    "SLEPC": eigs_slepc_spawn,
    "SLEPC-NOMPI": eigs_slepc,
}


def eigensystem_partial(
    A,
    k,
    isherm,
    *,
    B=None,
    which=None,
    return_vecs=True,
    sigma=None,
    ncv=None,
    tol=None,
    v0=None,
    sort=True,
    backend=None,
    fallback_to_scipy=False,
    **backend_opts,
):
    """Return a few eigenpairs from an operator.

    Parameters
    ----------
    A : sparse, dense or linear operator
        The operator to solve for.
    k : int
        Number of eigenpairs to return.
    isherm : bool
        Whether to use hermitian solve or not.
    B : sparse, dense or linear operator, optional
        If given, the RHS operator defining a generalized eigen problem.
    which : {'SA', 'LA', 'LM', 'SM', 'TR'}
        Where in spectrum to take eigenvalues from (see
        :func:``scipy.sparse.linalg.eigsh``)
    return_vecs : bool, optional
        Whether to return the eigenvectors.
    sigma : float, optional
        Which part of spectrum to target, implies which='TR' if which is None.
    ncv : int, optional
        number of lanczos vectors, can use to optimise speed
    tol : None or float
        Tolerance with which to find eigenvalues.
    v0 : None or 1D-array like
        An initial vector guess to iterate with.
    sort : bool, optional
        Whether to explicitly sort by ascending eigenvalue order.
    backend : {'AUTO', 'NUMPY', 'SCIPY',
               'LOBPCG', 'SLEPC', 'SLEPC-NOMPI'}, optional
        Which solver to use.
    fallback_to_scipy : bool, optional
        If an error occurs and scipy is not being used, try using scipy.
    backend_opts
        Supplied to the backend solver.

    Returns
    -------
    elk : (k,) array
        The ``k`` eigenvalues.
    evk : (d, k) array
        Array with ``k`` eigenvectors as columns if ``return_vecs``.
    """
    settings = {
        "k": k,
        "B": B,
        "which": (
            "SA"
            if (which is None) and (sigma is None)
            else "TR"
            if (which is None) and (sigma is not None)
            else which
        ),
        "return_vecs": return_vecs,
        "sigma": sigma,
        "isherm": isherm,
        "ncv": ncv,
        "sort": sort,
        "tol": tol,
        "v0": v0,
    }

    # Choose backend to perform the decompostion
    bkd = "AUTO" if backend is None else backend.upper()
    if bkd == "AUTO":
        bkd = choose_backend(A, k, sigma is not None, B=B)

    try:
        return _EIGS_METHODS[bkd](A, **settings, **backend_opts)

    # sometimes e.g. lobpcg fails, worth trying scipy
    except Exception as e:  # pragma: no cover
        if fallback_to_scipy and (bkd != "SCIPY"):
            warnings.warn(
                f"`eigensystem_partial` with backend '{bkd}' failed, trying "
                "again with scipy. Set ``fallback_to_scipy=False`` to avoid "
                "this and see the full error."
            )

            return eigs_scipy(A, **settings, **backend_opts)
        else:
            raise e


# --------------------------------------------------------------------------- #
#                        Full eigendecomposition                              #
# --------------------------------------------------------------------------- #


def eigensystem(A, isherm, *, k=-1, sort=True, return_vecs=True, **kwargs):
    """Find all or some eigenpairs of an operator.

    Parameters
    ----------
    A : operator
        The operator to decompose.
    isherm : bool
        Whether the operator is assumed to be hermitian or not.
    k : int, optional
        If negative, find all eigenpairs, else perform partial
        eigendecomposition and find ``k`` pairs. See
        :func:`~quimb.linalg.base_linalg.eigensystem_partial`.
    sort : bool, optional
        Whether to sort the eigenpairs in ascending eigenvalue order.
    kwargs
        Supplied to the backend function.

    Returns
    -------
    el : (k,) array
        Eigenvalues.
    ev : (d, k) array
        Corresponding eigenvectors as columns of array, such that
        ``ev @ diag(el) @ ev.H == A``.
    """
    if k < 0:
        return eig_numpy(
            A, isherm=isherm, sort=sort, return_vecs=return_vecs, **kwargs
        )

    return eigensystem_partial(
        A, k=k, isherm=isherm, sort=sort, return_vecs=return_vecs, **kwargs
    )


eig = functools.partial(eigensystem, isherm=False, return_vecs=True)
eigh = functools.partial(eigensystem, isherm=True, return_vecs=True)
eigvals = functools.partial(eigensystem, isherm=False, return_vecs=False)
eigvalsh = functools.partial(eigensystem, isherm=True, return_vecs=False)


@functools.wraps(eigensystem)
def eigenvectors(A, isherm, *, sort=True, **kwargs):
    return eigensystem(A, isherm=isherm, sort=sort, **kwargs)[1]


eigvecs = functools.partial(eigenvectors, isherm=False)
eigvecsh = functools.partial(eigenvectors, isherm=True)


def groundstate(ham, **kwargs):
    """Alias for finding lowest eigenvector only."""
    return eigvecsh(ham, k=1, which="SA", **kwargs)


def groundenergy(ham, **kwargs):
    """Alias for finding lowest eigenvalue only."""
    return eigvalsh(ham, k=1, which="SA", **kwargs)[0]


def bound_spectrum(A, backend="auto", **kwargs):
    """Return the smallest and largest eigenvalue of hermitian operator ``A``."""
    el_min = eigvalsh(A, k=1, which="SA", backend=backend, **kwargs)[0]
    el_max = eigvalsh(A, k=1, which="LA", backend=backend, **kwargs)[0]
    return el_min, el_max


def _rel_window_to_abs_window(el_min, el_max, w_0, w_sz=None):
    """Convert min/max eigenvalues and relative window to absolute values.

    Parameters
    ----------
    el_min : float
        Smallest eigenvalue.
    el_max : float
        Largest eigenvalue.
    w_0 : float [0.0 - 1.0]
        Relative window centre.
    w_sz : float, optional
        Relative window width.

    Returns
    -------
    l_0[, l_min, l_max]:
        Absolute value of centre of window, lower and upper intervals if a
        window size is specified.
    """
    el_range = el_max - el_min
    el_w_0 = el_min + w_0 * el_range
    if w_sz is not None:
        el_w_min = el_w_0 - w_sz * el_range / 2
        el_w_max = el_w_0 + w_sz * el_range / 2
        return el_w_0, el_w_min, el_w_max
    return el_w_0


def eigh_window(
    A,
    w_0,
    k,
    w_sz=None,
    backend="AUTO",
    return_vecs=True,
    offset_const=1 / 104729,
    **kwargs,
):
    """Return mid-spectrum eigenpairs from a hermitian operator.

    Parameters
    ----------
    A : (d, d) operator
        Operator to retrieve eigenpairs from.
    w_0 : float [0.0, 1.0]
        Relative window centre to retrieve eigenpairs from.
    k : int
        Target number of eigenpairs to retrieve.
    w_sz : float, optional
        Relative maximum window width within which to keep eigenpairs.
    backend : str, optional
        Which :func:`~quimb.eigh` backend to use.
    return_vecs : bool, optional
        Whether to return eigenvectors as well.
    offset_const : float, optional
        Small fudge factor (relative to window range) to avoid 1 / 0 issues.

    Returns
    -------
    el : (k,) array
        Eigenvalues around w_0.
    ev : (d, k) array
        The eigenvectors, if ``return_vecs=True``.
    """
    w_sz = w_sz if w_sz is not None else 1.1

    if isdense(A) or backend.upper() == "NUMPY":
        if return_vecs:
            lk, vk = eigh(A.toarray() if issparse(A) else A, **kwargs)
        else:
            lk = eigvalsh(A.toarray() if issparse(A) else A, **kwargs)

        lmin, lmax = lk[0], lk[-1]
        l_w0, l_wmin, l_wmax = _rel_window_to_abs_window(lmin, lmax, w_0, w_sz)

    else:
        lmin, lmax = bound_spectrum(A, backend=backend, **kwargs)
        l_w0, l_wmin, l_wmax = _rel_window_to_abs_window(lmin, lmax, w_0, w_sz)
        l_w0 += (lmax - lmin) * offset_const  # for 1/0 issues

        if return_vecs:
            lk, vk = eigh(A, k=k, sigma=l_w0, backend=backend, **kwargs)
        else:
            lk = eigvalsh(A, k=k, sigma=l_w0, backend=backend, **kwargs)

    # Trim eigenpairs from beyond window
    in_window = (lk > l_wmin) & (lk < l_wmax)

    if return_vecs:
        return lk[in_window], vk[:, in_window]

    return lk[in_window]


def eigvalsh_window(*args, **kwargs):
    """Alias for only finding the eigenvalues in a relative window."""
    return eigh_window(*args, return_vecs=False, **kwargs)


def eigvecsh_window(*args, **kwargs):
    """Alias for only finding the eigenvectors in a relative window."""
    return eigh_window(*args, return_vecs=True, **kwargs)[1]


# -------------------------------------------------------------------------- #
# Partial singular value decomposition                                       #
# -------------------------------------------------------------------------- #


def svd(A, return_vecs=True):
    """Compute full singular value decomposition of an operator, using numpy.

    Parameters
    ----------
    A : (m, n) array
        The operator.
    return_vecs : bool, optional
        Whether to return the singular vectors.

    Returns
    -------
    U : (m, k) array
        Left singular vectors (if ``return_vecs=True``) as columns.
    s : (k,) array
        Singular values.
    VH : (k, n) array
        Right singular vectors (if ``return_vecs=True``) as rows.
    """
    try:
        return np.linalg.svd(A, full_matrices=False, compute_uv=return_vecs)

    except np.linalg.LinAlgError:  # pragma: no cover
        warnings.warn("Numpy SVD failed, trying again with different driver.")
        return sla.svd(
            A,
            full_matrices=False,
            compute_uv=return_vecs,
            lapack_driver="gesvd",
        )


_SVDS_METHODS = {
    "SLEPC": svds_slepc_spawn,
    "SLEPC-NOMPI": svds_slepc,
    "NUMPY": svds_numpy,
    "SCIPY": svds_scipy,
    "PRIMME": svds_primme,
}


def svds(A, k, ncv=None, return_vecs=True, backend="AUTO", **kwargs):
    """Compute the partial singular value decomposition of an operator.

    Parameters
    ----------
    A : dense, sparse or linear operator
        The operator to decompose.
    k : int, optional
        number of singular value (triplets) to retrieve
    ncv : int, optional
        Number of lanczos vectors to use performing decomposition.
    return_vecs : bool, optional
        Whether to return the left and right vectors
    backend : {'AUTO', 'SCIPY', 'SLEPC', 'SLEPC-NOMPI', 'NUMPY'}, optional
        Which solver to use to perform decomposition.

    Returns
    -------
    (Uk,) sk (, VHk) :
        Singular value(s) (and vectors) such that ``Uk @ np.diag(sk) @ VHk``
        approximates ``A``.
    """
    settings = {"k": k, "ncv": ncv, "return_vecs": return_vecs}

    bkd = (
        choose_backend(A, k, False)
        if backend in {"auto", "AUTO"}
        else backend.upper()
    )
    svds_func = _SVDS_METHODS[bkd.upper()]

    return svds_func(A, **settings, **kwargs)


# -------------------------------------------------------------------------- #
# Norms and other quantities based on decompositions                         #
# -------------------------------------------------------------------------- #


def norm_2(A, **kwargs):
    """Return the 2-norm of operator, ``A``, i.e. the largest singular value."""
    return svds(A, k=1, return_vecs=False, **kwargs)[0]


def norm_fro_dense(A):
    """Frobenius norm for dense matrices"""
    return vdot(A, A).real ** 0.5


def norm_fro_sparse(A):
    return vdot(A.data, A.data).real ** 0.5


def norm_trace_dense(A, isherm=False):
    """Returns the trace norm of operator ``A``, that is,
    the sum of the absolute eigenvalues.
    """
    if isherm:
        return abs(eigvalsh(A)).sum()
    else:
        return svd(A, return_vecs=False).sum()


def norm(A, ntype=2, **kwargs):
    """Operator norms.

    Parameters
    ----------
    A : operator
        The operator to find norm of.
    ntype : str
        Norm to calculate, if any of:

        - {2, '2', 'spectral'}: largest singular value
        - {'f', 'fro'}: frobenius norm
        - {'t', 'nuc', 'tr', 'trace'}: sum of singular values

    Returns
    -------
    x : float
        The operator norm.
    """
    types = {
        "2": "2",
        2: "2",
        "spectral": "2",
        "f": "f",
        "fro": "f",
        "t": "t",
        "trace": "t",
        "nuc": "t",
        "tr": "t",
    }
    methods = {
        ("2", 0): norm_2,
        ("2", 1): norm_2,
        ("t", 0): norm_trace_dense,
        ("f", 0): norm_fro_dense,
        ("f", 1): norm_fro_sparse,
    }
    return methods[(types[ntype], issparse(A))](A, **kwargs)


# --------------------------------------------------------------------------- #
#                               Matrix functions                              #
# --------------------------------------------------------------------------- #


def expm(A, herm=False):
    """Matrix exponential, can be accelerated if explicitly hermitian.

    Parameters
    ----------
    A : dense or sparse operator
        Operator to exponentiate.
    herm : bool, optional
        If True (not default), and ``A`` is dense, digonalize the matrix
        in order to perform the exponential.
    """
    if issparse(A):
        # convert to and from csc to suppress scipy warning
        return spla.expm(A.tocsc()).tocsr()
    elif not herm:
        return qarray(spla.expm(A))
    else:
        evals, evecs = eigh(A)
        return evecs @ ldmul(np.exp(evals), dag(evecs))


_EXPM_MULTIPLY_METHODS = {
    "SCIPY": spla.expm_multiply,
    "SLEPC": functools.partial(mfn_multiply_slepc_spawn, fntype="exp"),
    "SLEPC-KRYLOV": functools.partial(
        mfn_multiply_slepc_spawn, fntype="exp", MFNType="KRYLOV"
    ),
    "SLEPC-EXPOKIT": functools.partial(
        mfn_multiply_slepc_spawn, fntype="exp", MFNType="EXPOKIT"
    ),
    "SLEPC-NOMPI": functools.partial(mfn_multiply_slepc, fntype="exp"),
}


def expm_multiply(mat, vec, backend="AUTO", **kwargs):
    """Compute the action of ``expm(mat)`` on ``vec``.

    Parameters
    ----------
    mat : operator
        Operator with which to act with exponential on ``vec``.
    vec : vector-like
        Vector to act with exponential of operator on.
    backend : {'AUTO', 'SCIPY', 'SLEPC', 'SLEPC-KRYLOV', 'SLEPC-EXPOKIT'}
        Which backend to use.
    kwargs
        Supplied to backend function.

    Returns
    -------
    vector
        Result of ``expm(mat) @ vec``.
    """
    if backend == "AUTO":
        if SLEPC4PY_FOUND and vec.size > 2**10:
            backend = "SLEPC"
        else:
            backend = "SCIPY"

    return _EXPM_MULTIPLY_METHODS[backend.upper()](mat, vec, **kwargs)


def sqrtm(A, herm=True):
    """Matrix square root, can be accelerated if explicitly hermitian.

    Parameters
    ----------
    A : dense array
        Operator to take square root of.
    herm : bool, optional
        If True (the default), and ``A`` is dense, digonalize the matrix
        in order to take the square root.

    Returns
    -------
    array
    """
    if issparse(A):
        raise NotImplementedError("No sparse sqrtm available.")
    elif not herm:
        return qarray(sla.sqrtm(A))
    else:
        evals, evecs = eigh(A)
        return evecs @ ldmul(np.sqrt(evals.astype(complex)), dag(evecs))


class IdentityLinearOperator(spla.LinearOperator):
    """Get a ``LinearOperator`` representation of the identity operator,
    scaled by ``factor``.

    Parameters
    ----------
    size : int
        The size of the identity.
    factor : float
        The coefficient of the identity.

    Examples
    --------

    >>> I3 = IdentityLinearOperator(100, 1/3)
    >>> p = rand_ket(100)
    >>> np.allclose(I3 @ p, p / 3)
    True
    """

    def __init__(self, size, factor=1):
        self.factor = factor
        super().__init__(dtype=np.array(factor).dtype, shape=(size, size))

    def _matvec(self, vec):
        return self.factor * vec

    def _rmatvec(self, vec):
        return self.factor * vec

    def _matmat(self, mat):
        return self.factor * mat


class Lazy:
    """A simple class representing an unconstructed matrix. This can be passed
    to, for example, MPI workers, who can then construct the matrix themselves.
    The main function ``fn`` should ideally take an ``ownership`` keyword to
    avoid forming every row.

    This is essentially like using ``functools.partial`` and assigning the
    ``shape`` attribute.

    Parameters
    ----------
    fn : callable
        A function that constructs an operator.
    shape :
        Shape of the constructed operator.
    args
        Supplied to ``fn``.
    kwargs
        Supplied to ``fn``.

    Returns
    -------
    Lazy : callable

    Examples
    --------
    Setup the lazy operator:

    >>> H_lazy = Lazy(ham_heis, n=10, shape=(2**10, 2**10), sparse=True)
    >>> H_lazy
    <Lazy(ham_heis, shape=(1024, 1024), dtype=None)>

    Build a matrix slice (usually done automatically by e.g. ``eigs``):

    >>> H_lazy(ownership=(256, 512))
    <256x1024 sparse matrix of type '<class 'numpy.float64'>'
            with 1664 stored elements in Compressed Sparse Row format>
    """

    def __init__(self, fn, *args, shape=None, factor=None, **kwargs):
        if shape is None:
            raise TypeError("`shape` must be specified.")
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self.shape = shape
        self.factor = factor
        self.dtype = None

    def __imul__(self, x):
        if self.factor is None:
            self.factor = x
        else:
            self.factor = self.factor * x

    def __mul__(self, x):
        if self.factor is not None:
            x = x * self.factor
        return Lazy(
            self.fn, *self.args, shape=self.shape, factor=x, **self.kwargs
        )

    def __rmul__(self, x):
        return self.__mul__(x)

    def __call__(self, **kwargs):
        A = self.fn(*self.args, **self.kwargs, **kwargs)

        # check if any prefactors have been set
        if self.factor is not None:
            # try inplace first
            try:
                A *= self.factor
            except (ValueError, TypeError):
                A = self.factor * A

        # helpful to store dtype once constructed
        self.dtype = A.dtype
        return A

    def __repr__(self):
        s = "<Lazy({}, shape={}{}{})>"

        s_dtype = f", dtype={self.dtype}" if self.dtype is not None else ""
        s_factor = f", factor={self.factor}" if self.factor is not None else ""

        return s.format(self.fn.__name__, self.shape, s_dtype, s_factor)
